import rasterio
import rasterio.plot
from rasterio.plot import show
import rasterio.mask
from rasterio.warp import calculate_default_transform, reproject, Resampling
from rasterio.merge import merge
from rasterio.windows import from_bounds
from rasterio.fill import fillnodata

import rioxarray
import xarray

import glob, os, warnings
import numpy as np
import math
import matplotlib.pyplot as plt
from matplotlib_scalebar.scalebar import ScaleBar
from matplotlib.colors import TwoSlopeNorm
from matplotlib import cm
import fiona
import shapely.geometry
import geopandas as gpd
import pandas as pd


""" Simple open a raster file """

def rOpen(geotiff, band=1, returnArray=True, returnRes=False, returnCrs=False):
    """
    Open a raster file as an array. Options to output raster resolution and coordinate system
        geotiff: input raster filename (e.g. 'raster.tif')
        band: the band to read. Defaults to 1 (int)
        returnArray: return the raster as a numpy array. Default is True (bool)
        returnRes: return the raster resolution; this returns a tuple. Default is False (bool)
        returnCrs: return the raster coordinate system. Default is False (bool)
    """
    vals = []
    if returnArray == True:
        array = rasterio.open(geotiff).read(band)
        vals.append(array)
    if returnRes == True:
        res = rasterio.open(geotiff).res
        vals.append(res)
    if returnCrs == True:
        crs = rasterio.open(geotiff).crs
        vals.append(crs)
    if len(vals) == 1:
        vals = vals[0] 
    return vals


""" Simple way to plot raster file """
    
def show_fig(image, title=None, color='Spectral', ctitle='', bounds=None, res=None, vmin=None, vcenter=None, vmax=None, savefig=False):
    """
	Plotting a raster file
		image: input array (array-like) (e.g. rasterio.open('raster.tif').read(1))
		title: figure title (str) (e.g. 'velocity plot')
		color: figure colormap (str) (e.g. 'Spectral', 'RdBu', etc)
		ctitle: colorbar title (str) (e.g. 'm/yr')
		bounds: raster plot bounds, input: (left, right, bottom, top) (floats) 
		res: raster pixel resolution (int or float)
		vmin, vmax: plot minimum and maximum (int or float)
		vcenter: center of colormap
		savefig: If true, figure is saved as a .jpg in the current directory with the filename being the title input (boolean)
    """
    fig, ax = plt.subplots(figsize=(12,6))
    if vcenter != None:
    	divnorm = TwoSlopeNorm(vmin=vmin, vcenter=vcenter, vmax=vmax)
    	c = ax.imshow(image, cmap=color, extent=bounds, norm=divnorm)
    else:
    	c = ax.imshow(image, cmap=color, extent=bounds, vmin=vmin, vmax=vmax)
    if res != None:
        ax.add_artist(ScaleBar(dx=res, units='m')) # add scalebar
        ax.tick_params(left=False, bottom=False, labelleft=False, labelbottom=False)
    fig.colorbar(c, label=ctitle)
    fig.suptitle(title)
    plt.show()
    if savefig == True:
    	fig.savefig(title + '.jpg', dpi=500) # to save the plot as a jpg image


""" Geoprocessing tools:

	Reproject shapefile
	Buffer shapefile
	Clip shapefile
	Reproject and clip geotiff
	Fill missing geotiff data
	Mosaic geotiff
	Save array as raster with identical properties
		
"""

def shpReprojection(shapefile, crs, dst='reprojected_shape.shp'):
    """
	Reproject a shapefile
		shapefile: input shapefile path (str) (e.g. 'input_shapefile.shp')
		crs: output coordinate system (str) (e.g. 'EPSG:4326')
		dst: output filename (str) (e.g. 'output_shapefile.shp')
    """
    src = gpd.read_file(shapefile)
    src = src.to_crs(crs)
    src.to_file(dst)


def shpBuffer(shapefile, dist=100, dst='buffered_shape.shp'):
    """
    Add a buffer around a shapefile. Note that the shapefile must be in a projected coordinate system (e.g. UTM)
        shapefile: input shapefile path (str) (e.g. 'input_shapefile.shp')
        dist: buffer distance, m (int, float) (500)
        dst: output filename (str) (e.g. 'output_shapefile.shp')
    """
    shape = gpd.read_file(shapefile)
    
    if shape.crs.is_geographic == True: 
        warnings.warn("The shapefile is in a geographic coordinate system. Reproject the shapefile and try again")
        return None

    # Create a buffer
    buffered_data = shape.buffer(dist)

    # Create a new GeoDataFrame with the buffered data
    buffered_gdf = gpd.GeoDataFrame(geometry=buffered_data, crs=shape.crs)
    buffered_gdf.to_file(dst) # save the buffered shapefile


def shpClip(geotiff, shapefile, destination, nan_val=0, fill=True, pad_size=0, crop=True):
    """
	Clip a geotiff with a shapefile
		geotiff: input raster to clip (str) (e.g. 'input.tif')
		shapefile: shapefile used for clipping (str) (e.g. 'input_shape.shp')
		destination: output clipped raster filename (str) (e.g. 'output.tif')
		nan_val: NaN value (int or float)
		fill: If True, the pixels outside the features will be set to nan_val. If False, the output array will contain the original pixel data, and only the mask will be based on shapes (boolean) 
		pad_size: width of padding in pixels (int or float)
		crop: whether to crop the raster to the extent of the shapes
    """
    
    with rasterio.open(geotiff) as src:
        # we need to make a temporary shapefile with the same crs as the geotiff 
        shpReprojection(shapefile, src.crs, dst='temp.shp')
        with fiona.open('temp.shp', 'r') as openshape:
            shapes = [feature['geometry'] for feature in openshape]

        out_image, out_transform = rasterio.mask.mask(
            src, 
            shapes, 
            crop=crop, 
            filled=fill, 
            nodata=nan_val, 
            pad=True, 
            pad_width=pad_size
        )
        
        out_image[np.isnan(out_image)] = 0
        kwargs = src.meta
        kwargs.update({'driver': 'GTiff',
                       'height': out_image.shape[1],
                       'width': out_image.shape[2],
                       'transform': out_transform})
        with rasterio.open(destination, 'w', **kwargs) as dst:
            dst.write(out_image)
            
    for f in glob.glob('temp.*'): # remove the 'temp.*' shapefiles
        os.remove(f)


def tifReprojectionResample(file, reprojected_tif, crs, res, interp, extent_file=None):
    """
	Reproject, resample, and/or clip raster extent
		file: input raster to resample, reproject, and/or clip (str) (e.g. 'input.tif')
		reprojected_tif: output filename (str) (e.g. 'output_raster.tif')
		crs: target raster coordinate system (str) (e.g. 'EPGS:4326')
		res: target raster resolution (int or float)
		interp: resampling interpolation method (Resampling method from rasterio.warp module)  (e.g. rasterio.warp.Resampling.cubic_spline)
		extent_file: raster file to match extent (str)    
    """

    if extent_file != None:
        tifReprojectionResample(file, 'temp_output.tif', crs, res, interp, extent_file=None)
        file = 'temp_output.tif'
    
    with rasterio.open(file) as src:
        if extent_file is None:
            # keep initial file bounds
            transform, width, height = calculate_default_transform(
                src.crs,
                crs,
                src.width,
                src.height,
                *src.bounds,
                resolution=res
            )
        else:
            # this readjusts the bounds to be that of extent_file
            dst = rasterio.open(extent_file)
            transform, width, height = calculate_default_transform(
                src.crs,
                crs,
                dst.width,
                dst.height,
                *dst.bounds,
                resolution=res
            )

        kwargs = src.meta.copy()
        kwargs.update({
            'crs': crs,
            'transform': transform,
            'width': width,
            'height': height
        })

        with rasterio.open(reprojected_tif, 'w', **kwargs) as dst:
            reproject(
                source=rasterio.band(src, 1),
                destination=rasterio.band(dst, 1),
                src_transform=src.transform,
                src_crs=src.crs,
                dst_transform=transform,
                dst_crs=crs,
                resampling=interp   # Resampling.cubic_spline
            )
            
        if extent_file != None:
            os.remove('temp_output.tif')


def fillHole(file, dest='output_filled.tif', dist=10, iters=1):
    """
	Fill hole in raster
		file: input raster to fill (str) (e.g. 'input.tif')
		dest: output filename (str) (e.g. 'output_raster.tif')
		dist: the maximum number of pixels to search in all directions to find values to interpolate from (int)
		iters: the number of 3x3 smoothing filter passes to run (int)
    """
    with rasterio.open(file) as src:
        profile = src.profile
        inputs = src.read(1)
        
        fillmask = inputs.copy() # fillnodata is applied where the mask=0
        fillmask[inputs>=0] = 1
        fillmask[fillmask!=1] = 0
        
        inputFilled = fillnodata(inputs, mask=fillmask, max_search_distance=dist, smoothing_iterations=iters)
        inputFilled[pd.isnull(inputFilled) == True] = 0

    with rasterio.open(dest, 'w', **profile) as dst:
        dst.write_band(1, inputFilled)


def mosaic_files(files, mosaic_output):
    """
	Create a mosiac of the obtained raster files. reproject files to the same crs if needed
		files: input rasters to mosaic (list of str) (e.g. ['input1.tif', 'input2.tif', 'input3.tif', 'input4.tif'])
		mosaic_output: the number of 3x3 smoothing filter passes to run (int)
    """
    src_files_to_mosaic = []
    dst_crs = rasterio.open(files[0]).crs
    for file in files:
        src = rasterio.open(file)   
        if src.crs != dst_crs: # NOTE: files must have the same crs
            raise 'Both datasets do not share a common coordinate system. One must be reprojected.'
        src_files_to_mosaic.append(src)
        
    mosaic, out_trans = merge(src_files_to_mosaic)
    out_meta = src.meta.copy()
    out_meta.update({'driver': 'GTiff',
                             'height': mosaic.shape[1],
                             'width': mosaic.shape[2],
                             'transform': out_trans})
    with rasterio.open(mosaic_output, 'w', **out_meta) as dest:
        dest.write(mosaic)


def rasterLike(array, destination, geotiff):
    """
	Save array as raster, assuming it is 'like' the raster (same res, crs, size, etc)
		array: input array to save as a raster (array-like)
		destination: output filename (str) (e.g. 'output_raster.tif')
		geotiff: raster (str) (e.g. 'raster.tif)
    """
    with rasterio.open(geotiff) as src:
        kwargs = src.meta.copy()
    with rasterio.open(destination, 'w', **kwargs) as dst:
        dst.write(array, 1)
        

def rasterMath(geotiff1, geotiff2, eqn='-', interp=Resampling.cubic_spline, outfile=None):
    '''
    Perform basic math with two raster files, regardless of coordinate system, resolution, and extent
        geotiff1: input raster filename. Note that the second raster will be reprojected and clipped based on this raster (e.g. 'raster1.tif')
        geotiff2: input raster filename (e.g. 'raster2.tif')
        eqn: raster math to perform; addition, subtraction, multiplication, or division. Add 'r' to reverse the order of input files for subtraction or division (to do geotiff2 - geotiff1 or geotiff2 / geotiff1) (e.g. '+', '-', '*', '/', '-r', '/r' OR 'add', 'sub', 'mult', 'div', 'subr', or 'divr')
        interp: resampling interpolation method (Resampling method from rasterio.warp module)  (e.g. rasterio.warp.Resampling.cubic_spline)
        outfile: output file to save raster. Output filename to save. If none, only the array is return (default: None) (str, e.g. 'output.tif')
    '''
    res = rOpen(geotiff1, returnArray=False, returnRes=True, returnCrs=False) # geotiff resolution
    crs = rOpen(geotiff1, returnArray=False, returnRes=False, returnCrs=True) # geotiff coordinate system
    ext = rasterio.open(geotiff1).bounds # geotiff bounds
    arr = rOpen(geotiff1) # geotiff array
    
    res2 = rOpen(geotiff2, returnArray=False, returnRes=True, returnCrs=False) # geotiff resolution
    crs2 = rOpen(geotiff2, returnArray=False, returnRes=False, returnCrs=True) # geotiff coordinate system
    ext2 = rasterio.open(geotiff2).bounds # geotiff bounds
    
    if res == res2 and crs == crs2 and ext == ext2:
        arr2 = rOpen(geotiff2) # if our geotiffs are aligned, we just take the second array
    else: 
        # otherwise we need to reproject and resample our array temporarily
        tifReprojectionResample(geotiff2, 'temp_math_reproj.tif', crs=crs, res=res, 
                                interp=interp, extent_file=geotiff1)
        arr2 = rOpen('temp_math_reproj.tif')
        os.remove('temp_math_reproj.tif')

    # perform the arithmetic
    if eqn == '-' or eqn == 'sub':
        math_arr = np.subtract(arr, arr2)
    elif eqn == '+' or eqn == 'add':
        math_arr = np.add(arr, arr2)
    elif eqn == '*' or eqn == 'mult':
        math_arr = np.multiply(arr, arr2)
    elif eqn == '/' or eqn == 'div':
        math_arr = np.divide(arr, arr2)
    elif eqn == '-r' or eqn == 'subr':
        math_arr = np.subtract(arr2, arr)
    elif eqn == '/r' or eqn == 'divr':
        math_arr = np.divide(arr2, arr)    
        
    if outfile != None: # save as geotiff
        rasterLike(math_arr, outfile, geotiff1)      
    return math_arr



""" Extra functions for sampling along lines """


def extract_along_line(xarr, line, n_samples=512):
    # samples line with n_samples number of points
    profile = []
    dist = []
    for i in range(n_samples):
        point = line.interpolate(i / n_samples - 1., normalized=True) # get next point on the line
        value = xarr.sel(x=point.x, y=point.y, method="nearest").data # access the nearest pixel in the xarray
        profile.append(value)
        dist.append([point.x, point.y])
    return profile, dist


def points_along_lines(geotiff, line, ID='SEGMENT_ID'):
    # extract raster values along the line
    # line is a shapely geometry: shapely.geometry.MultiLineString(...)
    # xarr is a rioxarray array: rioxarray.open_rasterio(geoTIFF).squeeze()
    xarr = rioxarray.open_rasterio(geotiff).squeeze()
    profile, coordinate = [], []
    ids = line[ID]
    for g in line['geometry']:
        prof, coord = extract_along_line(xarr, g)
        profile.append(prof)
        coordinate.append(coord)

    # get the raster values along the line and the distances along the transect
    vals, line_dists, dist = [], [], []
    for n in range(len(profile)):
        v = [float(x) for x in profile[n]]
        line_d = [[coordinate[n][i], coordinate[n][i+1]] for i in range(len(coordinate[n])-1)]
        d = [math.dist(pair[0], pair[1]) for pair in line_d]
        d = [sum(d[:i]) for i in range(len(d)+1)]
        vals.append(v)
        line_dists.append(line_d)
        dist.append(d)
    return vals, line_dists, dist, ids.to_numpy()


def end_points(xarr, points):
    # obtains raster values at our point selections themselves (not interpolated lines)
    point_spot = []
    for p in points:
        point = xarr.sel(x=p[0], y=p[1], method="nearest").data # access the nearest pixel in the xarray
        point_spot.append(point)
    return point_spot
    
def distance_to_shp(geotiff, gpd_shapefile, common_crs='EPSG:32606'):
    '''
    Gets the distance from every pixel to a shapefile
    :param geotiff: input raster file (geotiff as string)
    :param gpd_shapefile: shapefile object to calculate distance from, e.g. centerlines (gpd read shapefile)
    :param common_crs: common crs for geotiff and shapefile data, e.g. EPSG:32606 (coordinate as string)
    :return: array of values with distance from shapefile
    '''
    with rasterio.open(geotiff) as dataset:
        val = dataset.read(1)
        data = [(dataset.xy(x,y)[0],dataset.xy(x,y)[1],val[x,y]) if (np.isnan(val[x,y]) == False) 
                else (np.nan,np.nan,np.nan) for x,y in np.ndindex(val.shape)]
        lon = [i[0] for i in data]
        lat = [i[1] for i in data]
        d = [i[2] for i in data]
        res = pd.DataFrame({"lon":lon,'lat':lat,"data":d})
        res_geodf = gpd.GeoDataFrame(res, geometry=gpd.points_from_xy(res['lon'], res['lat'], crs=dataset.crs))

    # Convert both line and points to the same projected CRS.
    gpd_shapefile_com = gpd_shapefile.to_crs(common_crs)
    res_geodf_com = res_geodf.to_crs(common_crs)

    # The track line shapefile may contain several lines. To measure the shortest distance do a union of all tracks
    single_gpd_shapefile = gpd_shapefile_com.unary_union
    res_geodf_com['distance_m'] = res_geodf_com.distance(single_gpd_shapefile) # Measure the distance
    distances = res_geodf_com['distance_m'].to_numpy()
    distances_array = np.reshape(distances, val.shape)
    return distances_array






import matplotlib.pyplot as plt
import numpy as np
import earthpy.plot as ep
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.colors import TwoSlopeNorm, ListedColormap, BoundaryNorm
from matplotlib_scalebar.scalebar import ScaleBar


def show_fig(image, title=None, color='Spectral', ctitle='', bounds=None, res=None, vmin=None, vcenter=None, vmax=None, savefig=False):
    """
	Plotting a raster file
		image: input array (array-like) (e.g. rasterio.open('raster.tif').read(1))
		title: figure title (str) (e.g. 'velocity plot')
		color: figure colormap (str) (e.g. 'Spectral', 'RdBu', etc)
		ctitle: colorbar title (str) (e.g. 'm/yr')
		bounds: raster plot bounds, input: (left, right, bottom, top) (floats) 
		res: raster pixel resolution (int or float)
		vmin, vmax: plot minimum and maximum (int or float)
		vcenter: center of colormap
		savefig: If true, figure is saved as a .jpg in the current directory with the filename being the title input (boolean)
    """
    fig, ax = plt.subplots(figsize=(12,6))
    if vcenter != None:
    	divnorm = TwoSlopeNorm(vmin=vmin, vcenter=vcenter, vmax=vmax)
    	c = ax.imshow(image, cmap=color, extent=bounds, norm=divnorm)
    else:
    	c = ax.imshow(image, cmap=color, extent=bounds, vmin=vmin, vmax=vmax)
    if res != None:
        ax.add_artist(ScaleBar(dx=res, units='m')) # add scalebar
    fig.colorbar(c, label=ctitle)
    fig.suptitle(title)
    plt.show()
    if savefig == True:
    	fig.savefig(title + '.jpg', dpi=500) # to save the plot as a jpg image

def scatterplot(data, xtitle, ytitle, title, colors, labels, markers='.', alpha=0.7, xyLine=False, buff=100, savefig=False):
    """
	Create a scatterplot from data
		data: input data (list of arrays of length 2) (e.g. [[0, 0, 0, 0], [1, 2, 3, 4]] or [[[0, 0, 0], [1, 2, 3]], [[1, 1, 1, 1], [1, 2, 3, 4]]])
		xtitle, ytitle: axis titles (str)
		colors: list of colors for each x-y scatterplot dataset (list of str, same length as data)
		labels: legend label for data (list of str, same length as data) (e.g. ['label1', 'label2'])
		markers: scatterplot marker (list of str, same length as data, or string if all markers should be the same)
		alpha: marker alpha value; transparency (float ranging 0 to 1)
		xyLine: to show the 1-to-1 line (boolean)
		buff: buffer on axis limits (int or float)
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)
    plt.grid()
    
    if type(markers) == str:
        markers = [markers] * len(data)

    for i, xy in enumerate(data):
        ax.scatter(xy[0], xy[1], c=colors[i], alpha=alpha, marker=markers[i], label=labels[i])
    ax.set_xlabel(xtitle)
    ax.set_ylabel(ytitle)
    ax.legend()
    if xyLine == True:
        ax.axline([0, 0], [1, 1], c='k', ls='--', lw=1)

    maximums_x, minimums_x, maximums_y, minimums_y = [], [], [], []
    for d in np.array(data):
        maximums_x.append(np.nanmax(d[0]))
        minimums_x.append(np.nanmin(d[0]))
        maximums_y.append(np.nanmax(d[1]))
        minimums_y.append(np.nanmin(d[1]))

    ax.set_xlim([min(minimums_x) - buff, max(maximums_x) + buff])
    ax.set_ylim([min(minimums_y) - buff, max(maximums_y) + buff])
    fig.suptitle(title, size=8)
    
    if savefig == True:
    	figName = title.replace(' ', '_') + '.png'
    	fig.savefig(figName, dpi=250)


def plotMany(cbarTitle, color, plotTitle, *kwargs):
    """
	Create a figure with many plots
		cbarTitle: colorbar title (str)
		color: colormap (str) (e.g. 'Spectral', 'RdBu', 'gist_earth')
		plotTitle: title of the figure (str)
	
		All *kwargs should follow this: [data, title, v] where v is [vmin, vcenter, vmax]. (list of lists, on per plot)
			data: array of data (array-like)
			title: plot title (str)
			v: minimum, center, and maximum value for plot colormap
    """
    # plt.subplots([rows of plots], [columns of plots], figsize=([width], [height])
    l = len(kwargs[0]) + 3
    row, col = (int(l/4), 4)
    fig, ax = plt.subplots(row, col, figsize=(3*col, 3*row), facecolor='w', edgecolor='k')
    # fig.subplots_adjust(hspace = .5, wspace=.001)
    ax = ax.ravel()

    for i, arg in enumerate(kwargs[0]):
        data, title, v = arg
        divnorm = TwoSlopeNorm(vmin=v[0], vcenter=v[1], vmax=v[2])
        im = ax[i].imshow(data, cmap=plt.cm.get_cmap(color), norm=divnorm)
        divider = make_axes_locatable(ax[i])
        cax = divider.append_axes('right', size='5%', pad=0.05)
        fig.colorbar(im, cax=cax, label=cbarTitle)
        ax[i].set_title(title)

    fig.tight_layout()
    fig.suptitle(plotTitle, weight='bold')
    plt.show()
    figName = plotTitle.replace(' ', '_') + '.png'
    # fig.savefig(figName, dpi=250)

    # EXAMPLE:
    # icepack_v_list = [np.array(np.sqrt(vx ** 2 + vy ** 2)) for vx, vy in zip(icepack_vx_list, icepack_vy_list)]
    # icepack_title_list = ['Icepack Velocity (n=' + str(n + 1) + ')' for n in range(l)]
    # minmax_list = [[0, 5, 60] for i in icepack_vx_list]
    # plotMany('Velocity (m/yr)', 'BrBG', 'Plot of All IcePack Results', list(zip(icepack_v_list, icepack_title_list, minmax_list)))


def elevationBinPlot(xVal1, yVal1, xVal2, yVal2, xLabel1, yLabel1, xLabel2, title1,
                     elevationBinWidth, minVal, buff, alpha, savefig=False):
    '''
    Plot y (elevation bin) vs two x data sources: one primary one (xVal1) and a secondary bar graph (xVal2)
    :param xVal1: Primary x values (scatter plot)
    :param yVal1: Primary values
    :param xVal2: Secondary x values (bar graph)
    :param yVal2: Secondary y values
    :param xLabel1, xlabel2, yLabel1, title: Chart axis labels and titles (xLabel2 is above plot)
    :param elevationBinWidth: Elevation bin width
    :param minVal: Minimum x-value of plot: if 0, plot begins at 0. Otherwise it begins at the minimum value in
    xVal1 minus buff.
    :param buff: x-axis minimum and maximum value buffer
    :param alpha: opacity of bar plot
    :return: No return: plots the figure
    '''
    fig = plt.figure()
    ax = fig.add_subplot(111, label="1")
    ax2 = fig.add_subplot(131, label="2", frame_on=False)

    ax.scatter(xVal1, yVal1)
    ax.set_xlabel(xLabel1)
    ax.set_ylabel(yLabel1)
    ax.tick_params(axis='x')
    ax.barh(xVal2, yVal2, elevationBinWidth, alpha=0)
    if minVal == 0:
        ax.set_xlim([0, max(xVal1)+buff])
    else:
        ax.set_xlim([min(xVal1) - buff, max(xVal1) + buff])

    ax2.barh(xVal2, yVal2, elevationBinWidth, alpha=alpha, color="C1", zorder=0)
    ax2.xaxis.tick_top()
    ax2.set_xlabel(xLabel2, color="C1")
    ax2.xaxis.set_label_position('top')
    ax2.tick_params(axis='x', colors="C1")

    fig.suptitle(title1, weight='bold')
    fig.tight_layout()
    
    if savefig == True:
    	plt.show(block=False)
    	figName = title1.replace(' ', '_') + '.png'
    	fig.savefig(figName, dpi=250)
    	# Image.open(figName).show()
    else:
    	plt.show()

def elevationBinPlot2(xVal1, labx1, xVal1_2, labx1_2, yVal1, xVal2, yVal2, xLabel1, yLabel1, xLabel2, title1,
                     elevationBinWidth, minVal, buff, alpha, savefig=False):
    '''
    Plot y (elevation bin) vs two x data sources: one primary one (xVal1) and a secondary bar graph (xVal2)
    :param xVal1: Primary x values (scatter plot)
    :param labx1: scatter plot legend label for xVal1
    :param xVal1_2: Second x values (scatter plot)
    :param labx1_2: scatter plot legend label for xVal1_2
    :param yVal1: Primary values
    :param xVal2: Secondary x values (bar graph)
    :param yVal2: Secondary y values
    :param xLabel1, xlabel2, yLabel1, title: Chart axis labels and titles (xLabel2 is above plot)
    :param elevationBinWidth: Elevation bin width
    :param minVal: Minimum x-value of plot: if 0, plot begins at 0. Otherwise it begins at the minimum value in
    xVal1 minus buff.
    :param buff: x-axis minimum and maximum value buffer
    :param alpha: opacity of bar plot
    '''
    fig = plt.figure()
    ax = fig.add_subplot(111, label="1")
    ax2 = fig.add_subplot(131, label="2", frame_on=False)

    ax.scatter(xVal1, yVal1, label=labx1)
    ax.set_xlabel(xLabel1)
    ax.set_ylabel(yLabel1)
    ax.tick_params(axis='x')
    ax.barh(xVal2, yVal2, elevationBinWidth, alpha=0, zorder=0)
    if minVal == 0:
        ax.set_xlim([0, max(xVal1)+buff])
    else:
        ax.set_xlim([min(xVal1) - buff, max(xVal1) + buff])

    ax.scatter(xVal1_2, yVal1, label=labx1_2)
    ax.legend()

    ax2.barh(xVal2, yVal2, elevationBinWidth, alpha=alpha, color="C1")
    ax2.xaxis.tick_top()
    ax2.set_xlabel(xLabel2, color="C1")
    ax2.xaxis.set_label_position('top')
    ax2.tick_params(axis='x', colors="C1")

    fig.suptitle(title1, weight='bold')
    fig.tight_layout()
    if savefig == True:
    	plt.show(block=False)
    	figName = title1.replace(' ', '_') + '.png'
    	fig.savefig(figName, dpi=250)
    	# Image.open(figName).show()
    else:
    	plt.show()

def elevationBinPlot3Subfigs(x_subp1, x_subp1_lab, x_subp1_2, x_subp1_lab_2, x_subp1_3, x_subp1_lab_3, y_subp1,
                             x2_subp1, y2_subp1, xLabel1, yLabel1, xLabel2, title1, elBinWidth, buff, alpha, title, res,
                             subp3, title3, cbar3, color3,
                             subp4, title4, cbar4, color4,
                             subp5, title5, cbar5, color5,
                             err_subp1_2=None, err_subp1_3=None,
                             x_subp1_ref=None, x_subp1_lab_ref=None, y_subp1_ref=None, ref_var=None, cbar4ticks=None, savefig=False):
    '''
    Plot y (elevation bin) vs two x data sources: one primary one (xVal1) and a secondary bar graph (xVal2)
    Includes 3 subplots showing a desired source: emergence velocity, thickness, speed
    :param x_subp1: Primary x values (scatter plot)
    :param x_subp1_lab: scatter plot legend label for xVal1
    :param x_subp1_2: Second x values (scatter plot)
    :param x_subp1_lab_2: scatter plot legend label for xVal1_2
    :param x_subp1_3: Third x values (scatter plot)
    :param x_subp1_lab_3: scatter plot legend label for xVal1_3
    :param x_subp1_ref: x-values for the known data
    :param x_subp1_lab_ref: label for the known data
    :param y_subp1: Primary values
    :param y_subp1_ref: y-values for the known data
    :param x2_subp1: Secondary x values (bar graph)
    :param y2_subp1: Secondary y values
    :param xLabel1, xlabel2, yLabel1, title: Chart axis labels and titles (xLabel2 is above plot)
    :param title1: title of subplot 1
    :param elevationBinWidth: Elevation bin width
    :param buff: x-axis minimum and maximum value buffer
    :param alpha: opacity of bar plot
    :return: No return: plots the figure
    :param title: overall figure title
    :param subp2, subp3, subp4: array for subplots 2, 3, and 4
    :param title2, title3, title4: title for subplots 2, 3, and 4
    :param cbar2, cbar3, cbar4: colorbar for subplots 2, 3, and 4
    '''
    fig = plt.figure()
    ax = fig.add_subplot(221, label="1")
    ax2 = fig.add_subplot(261, label="2", frame_on=False)

    ax3 = fig.add_subplot(222)
    ax4 = fig.add_subplot(223)
    ax5 = fig.add_subplot(224)

    ax.scatter(x_subp1, y_subp1, label=x_subp1_lab, color='w', edgecolor='dimgray', zorder=2)
    ax.plot(x_subp1, y_subp1, color='dimgray', zorder=1)
    ax.set_xlabel(xLabel1)
    ax.set_ylabel(yLabel1)
    ax.set_title(title1, pad=10)
    ax.tick_params(axis='x')
    ax.barh(x2_subp1, y2_subp1, elBinWidth, alpha=0, zorder=0)
    if err_subp1_2 is not None:
        if x_subp1_ref is None:
            if len(err_subp1_2.shape) > 1:
                ax.set_xlim([min(np.min(x_subp1_2 - err_subp1_2[0]), np.min(x_subp1_3 - err_subp1_3[0])) - 50,
                             max(np.max(x_subp1_2 + err_subp1_2[-1]), np.max(x_subp1_3 + err_subp1_3[-1])) + 50])
            else:
                ax.set_xlim([min(np.min(x_subp1_2 - err_subp1_2), np.min(x_subp1_3 - err_subp1_3)) - 50,
                             max(np.max(x_subp1_2 + err_subp1_2), np.max(x_subp1_3 + err_subp1_3)) + 50])
        else:
            if len(err_subp1_2.shape) > 1:
                ax.set_xlim([min(np.min(x_subp1_2 - err_subp1_2[0]), np.min(x_subp1_3 - err_subp1_3[0]),
                                 np.min(x_subp1_ref)) - 20,
                             max(np.max(x_subp1_2 + err_subp1_2[-1]), np.max(x_subp1_3 + err_subp1_3[-1]),
                                 np.max(x_subp1_ref), 0) + 20])
            else:
                ax.set_xlim(
                    [min(np.min(x_subp1_2 - err_subp1_2), np.min(x_subp1_3 - err_subp1_3), np.min(x_subp1_ref)) - 20,
                     max(np.max(x_subp1_2 + err_subp1_2), np.max(x_subp1_3 + err_subp1_3), np.max(x_subp1_ref),
                         0) + 20])
            ax.plot(x_subp1_ref, y_subp1_ref, color='black', zorder=5)  # plot line to connect reference data
    elif x_subp1_ref is not None:
        ax.set_xlim([min(np.min(x_subp1_ref), np.min(x_subp1))-buff, max(np.max(x_subp1_ref), np.max(x_subp1), 0)+buff])
        ax.plot(x_subp1_ref, y_subp1_ref, color='black', zorder=5)  # plot line to connect reference data
    else:
        ax.set_xlim([min(x_subp1) - buff, max(np.max(x_subp1), 0) + buff])

    ax2.barh(x2_subp1, y2_subp1, elBinWidth, alpha=alpha, color='dimgray')
    ax2.xaxis.tick_top()
    ax2.set_xlabel(xLabel2, color='dimgray')
    ax2.xaxis.set_label_position('top')
    ax2.tick_params(axis='x', colors='dimgray')

    ax.vlines(x_subp1_2, y_subp1 - np.random.uniform(elBinWidth/2, elBinWidth/2, len(y_subp1)),
              y_subp1 + np.random.uniform(elBinWidth/2, elBinWidth/2, len(y_subp1)), label=x_subp1_lab_2,
              lw=2, color='#01665e', zorder=7)
    ax.errorbar(x_subp1_2, y_subp1 - np.random.uniform(elBinWidth/10, elBinWidth/10, len(y_subp1)), xerr=err_subp1_2,
                fmt='', color='#01665e', ls='none', ecolor='#35978f', elinewidth=2, alpha=0.6, zorder=3)
    ax.vlines(x_subp1_3, y_subp1 - np.random.uniform(elBinWidth/2, elBinWidth/2, len(y_subp1)),
              y_subp1 + np.random.uniform(elBinWidth/2, elBinWidth/2, len(y_subp1)), label=x_subp1_lab_3,
              lw=2, color='#8c510a', zorder=6)
    ax.errorbar(x_subp1_3, y_subp1 + np.random.uniform(elBinWidth/10, elBinWidth/10, len(y_subp1)), xerr=err_subp1_3,
                fmt='', color='#8c510a', ls='none', ecolor='#bf812d', elinewidth=2, alpha=0.6, zorder=4)
    ax.scatter(x_subp1_ref, y_subp1_ref, label=x_subp1_lab_ref, color='black', zorder=5)
    ax.errorbar(x_subp1_ref, y_subp1_ref, xerr=ref_var,
                fmt='', color='black', ls='none', ecolor='black', elinewidth=2, alpha=0.6, zorder=3)
    if x_subp1_ref != None:
        m, b = np.polyfit(x_subp1_ref, y_subp1_ref, 1)
        ax.axline((0, b), slope=m, color='black', ls='--', alpha=0.8, zorder=6)
    ax.axvline(x=0, color='black', lw=0.5, dashes=(10, 3))
    ax.legend()

    # add 3 subplots
    # top right plot:
    divnorm3 = TwoSlopeNorm(vmin=min(subp3.min(), -1), vcenter=0, vmax=max(subp3.max(), 1))       # center colorbar at 0
    im3 = ax3.imshow(subp3, cmap=plt.cm.get_cmap(color3, len(x_subp1)), norm=divnorm3)   # plot array
    divider3 = make_axes_locatable(ax3)                         # plot on ax3 (northeast subplot)
    cax3 = divider3.append_axes('right', size='5%', pad=0.05)   # specify colorbar axis properties
    fig.colorbar(im3, cax=cax3, label=cbar3)                    # add colorbar
    ax3.set_title(title3, pad=10)                               # add subplot title
    ax3.add_artist(ScaleBar(dx=res, units='m'))

    # bottom left plot:
    divnorm4 = TwoSlopeNorm(vmin=min(subp4.min(), -1), vcenter=0, vmax=max(subp4.max(), 1))  # center colorbar at 0
    # subp4 = subp4.astype(float)
    # subp4[subp4 == 0] = np.nan                                  # to remove 0 values for white background in plot
    im4 = ax4.imshow(subp4, cmap=plt.cm.get_cmap(color4, len(x_subp1)), norm=divnorm4)
    divider4 = make_axes_locatable(ax4)
    cax4 = divider4.append_axes('right', size='5%', pad=0.05)
    cbar4 = fig.colorbar(im4, cax=cax4, label=cbar4)
    if cbar4ticks != None:              # this would be if we want colorbar ticks/labels different from actual values
        cbar4.ax.locator_params(nbins=len(cbar4ticks)//2)
        cbar4.ax.set_yticklabels(np.linspace(cbar4ticks[0], cbar4ticks[-1], len(cbar4ticks)//2).astype(int))
    ax4.set_title(title4, pad=10)
    ax4.add_artist(ScaleBar(dx=res, units='m'))

    # bottom right plot:
    divnorm5 = TwoSlopeNorm(vmin=min(subp5.min(), -1), vcenter=0, vmax=max(subp5.max(), 1))  # center colorbar at 0
    im5 = ax5.imshow(subp5, cmap=plt.cm.get_cmap(color5, len(x_subp1)), norm=divnorm5)
    divider5 = make_axes_locatable(ax5)
    cax5 = divider5.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im5, cax=cax5, label=cbar5)
    ax5.set_title(title5, pad=10)
    ax5.add_artist(ScaleBar(dx=res, units='m'))

    fig.set_size_inches(12, 8)
    fig.tight_layout(pad=0.5, w_pad=-3.0, h_pad=2.0)
    fig.suptitle(title, weight='bold')
    if savefig == True:
    	plt.show(block=False)
    	figName = title.replace(' ', '_') + '.png'
    	fig.savefig(figName, dpi=250)
    	# Image.open(figName).show()
    else:
    	plt.show()

def plotData(dataVals, cbarTitle, color, plotTitle, cluster=None, quiver=None, colorbarMin=None, colorbarMean=None, colorbarMax=None, savefig=False):
    '''
    Plots a map from input values (array-like)
    :param dataVals: values to be plotted
    :param cbarTitle: title of the colorbar
    :param color: colorbar scale (https://matplotlib.org/stable/tutorials/colors/colormaps.html)
    :param plotTitle: title of the plot
    :param cluster: number of discrete colorbar clusters. None by default, which has a continuous colorbar
    :param quiver: for quiver plot of arrows, list with 6 inputs from velPlot function below
    :param colorbarMax: maximum value of colorbar
    :return:
    '''
    fig = plt.figure()
    ax = fig.add_subplot(111, label="1")

    vmin = dataVals.min()
    vmean = dataVals.mean()
    vmax = dataVals.max()
    
    if colorbarMin != None:
        vmin = colorbarMin
    if colorbarMean != None:
        vmean = colorbarMean
    if colorbarMax != None:
        vmax = colorbarMax
    
    divnorm = TwoSlopeNorm(vmin=vmin, vcenter=vmean, vmax=vmax)
    dataVals = dataVals.astype(float)
    dataVals[dataVals == 0] = np.nan                                  # to remove 0 values for white background in plot
    im = ax.imshow(dataVals, cmap=plt.cm.get_cmap(color, cluster), norm=divnorm) #check len datavals
    if quiver != None:
        ax.quiver(quiver[0], quiver[1], quiver[2], quiver[3], quiver[4],
                   cmap=quiver[5], scale=quiver[6], width=.003)                   # velocity arrows
    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im, cax=cax, label=cbarTitle)
    ax.set_title(plotTitle, weight='bold', pad=10)
    fig.tight_layout(pad=3, w_pad=-3.0, h_pad=2.0)

    if savefig == True:
    	plt.show(block=False)
    	figName = plotTitle.replace(' ', '_') + '.png'
    	fig.savefig(figName, dpi=250)
    	# Image.open(figName).show()
    else:
    	plt.show()

def plotData3(dataVals1, cbarTitle1, color1, title1, dataVals2, cbarTitle2, color2, title2,
              dataVals3, cbarTitle3, color3, title3, plotTitle, res, quiver2=None, quiver3=None,
              div_select=[0,0,0], min=[0,0,0], mean=[50,50,50], max=[100,100,100], savefig=False):
    # div_select is to specify divnorm colorbar values for each of 3 plots or just have them be the data min, mean, max
    fig = plt.figure()
    ax1 = fig.add_subplot(131, label="1")
    ax2 = fig.add_subplot(132)
    ax3 = fig.add_subplot(133)

    if dataVals1.min() >= dataVals1.mean() or dataVals1.mean() >= dataVals1.max() or div_select[0] == 1:
        divnorm1 = TwoSlopeNorm(vmin=min[0], vcenter=mean[0], vmax=max[0])
    else:
        divnorm1 = TwoSlopeNorm(vmin=dataVals1.min(), vcenter=dataVals1.mean(), vmax=dataVals1.max())
        # divnorm1 = TwoSlopeNorm(vmin=750, vcenter=1000, vmax=1450)
    dataVals1 = dataVals1.astype(float)
    dataVals1[dataVals1 == 0] = np.nan  # to remove 0 values for white background in plot
    im1 = ax1.imshow(dataVals1, cmap=plt.cm.get_cmap(color1), norm=divnorm1)
    divider1 = make_axes_locatable(ax1)
    cax1 = divider1.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im1, cax=cax1, label=cbarTitle1)
    ax1.set_title(title1)
    ax1.add_artist(ScaleBar(dx=res, units='m'))

    if dataVals2.min() >= dataVals2.mean() or dataVals2.mean() >= dataVals2.max() or div_select[1] == 1:
        divnorm2 = TwoSlopeNorm(vmin=min[1], vcenter=mean[1], vmax=max[1])
    else:
        divnorm2 = TwoSlopeNorm(vmin=dataVals2.min(), vcenter=dataVals2.mean(), vmax=dataVals2.max())
    dataVals2 = dataVals2.astype(float)
    dataVals2[dataVals2 == 0] = np.nan  # to remove 0 values for white background in plot
    im2 = ax2.imshow(dataVals2, cmap=plt.cm.get_cmap(color2), norm=divnorm2)
    if quiver2 != None:
        ax2.quiver(quiver2[0], quiver2[1], quiver2[2], quiver2[3], quiver2[4],
                   cmap=quiver2[5], scale=quiver2[6], width=.003)                   # velocity arrows
    divider2 = make_axes_locatable(ax2)
    cax2 = divider2.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im2, cax=cax2, label=cbarTitle2)
    ax2.set_title(title2)
    ax2.add_artist(ScaleBar(dx=res, units='m'))

    if dataVals3.min() >= dataVals3.mean() or dataVals3.mean() >= dataVals3.max() or div_select[2] == 1:
        divnorm3 = TwoSlopeNorm(vmin=min[2], vcenter=mean[2], vmax=max[2])
    else:
        divnorm3 = TwoSlopeNorm(vmin=dataVals3.min(), vcenter=dataVals3.mean(), vmax=dataVals3.max())
        # divnorm3 = TwoSlopeNorm(vmin=dataVals3.min(), vcenter=dataVals3.mean(), vmax=50)
    dataVals3 = dataVals3.astype(float)
    dataVals3[dataVals3 == 0] = np.nan  # to remove 0 values for white background in plot
    im3 = ax3.imshow(dataVals3, cmap=plt.cm.get_cmap(color3), norm=divnorm3)
    if quiver3 != None:
        ax3.quiver(quiver3[0], quiver3[1], quiver3[2], quiver3[3], quiver3[4],
                   cmap=quiver3[5], scale=quiver3[6], width=.003)                   # velocity arrows
    divider3 = make_axes_locatable(ax3)
    cax3 = divider3.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im3, cax=cax3, label=cbarTitle3)
    ax3.set_title(title3)
    ax3.add_artist(ScaleBar(dx=res, units='m'))

    fig.set_size_inches(14, 4)
    fig.suptitle(plotTitle, weight='bold')
    fig.tight_layout(pad=3, w_pad=2.0, h_pad=0.0)
    if savefig == True:
    	plt.show(block=False)
    	figName = plotTitle.replace(' ', '_') + '.png'
    	fig.savefig(figName, dpi=250)
    	# Image.open(figName).show()
    else:
    	plt.show()

def plotData6(dataVals1, cbarTitle1, color1, title1, dataVals2, cbarTitle2, color2, title2,
              dataVals3, cbarTitle3, color3, title3, dataVals4, cbarTitle4, color4, title4,
              dataVals5, cbarTitle5, color5, title5, dataVals6, cbarTitle6, color6, title6, plotTitle, res, savefig=False):
    fig = plt.figure()
    ax1 = fig.add_subplot(231, label="1")
    ax2 = fig.add_subplot(232)
    ax3 = fig.add_subplot(233)
    ax4 = fig.add_subplot(234)
    ax5 = fig.add_subplot(235)
    ax6 = fig.add_subplot(236)

    divnorm1 = TwoSlopeNorm(vmin=dataVals1.min(), vcenter=dataVals1.mean(), vmax=dataVals1.max())
    # divnorm1 = TwoSlopeNorm(vmin=dataVals1.min(), vcenter=dataVals1.mean(), vmax=350)
    dataVals1 = dataVals1.astype(float)
    dataVals1[dataVals1 == 0] = np.nan  # to remove 0 values for white background in plot
    im1 = ax1.imshow(dataVals1, cmap=plt.cm.get_cmap(color1), norm=divnorm1)
    divider1 = make_axes_locatable(ax1)
    cax1 = divider1.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im1, cax=cax1, label=cbarTitle1)
    ax1.set_title(title1)
    ax1.add_artist(ScaleBar(dx=res, units='m'))

    divnorm2 = TwoSlopeNorm(vmin=dataVals2.min(), vcenter=dataVals2.mean(), vmax=dataVals2.max())
    dataVals2 = dataVals2.astype(float)
    dataVals2[dataVals2 == 0] = np.nan  # to remove 0 values for white background in plot
    im2 = ax2.imshow(dataVals2, cmap=plt.cm.get_cmap(color2), norm=divnorm2)
    divider2 = make_axes_locatable(ax2)
    cax2 = divider2.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im2, cax=cax2, label=cbarTitle2)
    ax2.set_title(title2)
    ax2.add_artist(ScaleBar(dx=res, units='m'))

    divnorm3 = TwoSlopeNorm(vmin=dataVals3.min(), vcenter=dataVals3.mean(), vmax=dataVals3.max())
    dataVals3 = dataVals3.astype(float)
    dataVals3[dataVals3 == 0] = np.nan  # to remove 0 values for white background in plot
    im3 = ax3.imshow(dataVals3, cmap=plt.cm.get_cmap(color3), norm=divnorm3)
    divider3 = make_axes_locatable(ax3)
    cax3 = divider3.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im3, cax=cax3, label=cbarTitle3)
    ax3.set_title(title3)
    ax3.add_artist(ScaleBar(dx=res, units='m'))

    divnorm4 = TwoSlopeNorm(vmin=dataVals4.min(), vcenter=dataVals4.mean(), vmax=dataVals4.max())
    # divnorm4 = TwoSlopeNorm(vmin=dataVals4.min(), vcenter=dataVals4.mean(), vmax=300)
    dataVals4 = dataVals4.astype(float)
    dataVals4[dataVals4 == 0] = np.nan  # to remove 0 values for white background in plot
    im4 = ax4.imshow(dataVals4, cmap=plt.cm.get_cmap(color4), norm=divnorm4)
    divider4 = make_axes_locatable(ax4)
    cax4 = divider4.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im4, cax=cax4, label=cbarTitle4)
    ax4.set_title(title4)
    ax4.add_artist(ScaleBar(dx=res, units='m'))

    divnorm5 = TwoSlopeNorm(vmin=dataVals5.min(), vcenter=dataVals5.mean(), vmax=dataVals5.max())
    dataVals5 = dataVals5.astype(float)
    dataVals5[dataVals5 == 0] = np.nan  # to remove 0 values for white background in plot
    im5 = ax5.imshow(dataVals5, cmap=plt.cm.get_cmap(color5), norm=divnorm5)
    divider5 = make_axes_locatable(ax5)
    cax5 = divider5.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im5, cax=cax5, label=cbarTitle5)
    ax5.set_title(title5)
    ax5.add_artist(ScaleBar(dx=res, units='m'))

    divnorm6 = TwoSlopeNorm(vmin=dataVals6.min(), vcenter=dataVals6.mean(), vmax=dataVals6.max())
    dataVals6 = dataVals6.astype(float)
    dataVals6[dataVals6 == 0] = np.nan  # to remove 0 values for white background in plot
    im6 = ax6.imshow(dataVals6, cmap=plt.cm.get_cmap(color6), norm=divnorm6)
    divider6 = make_axes_locatable(ax6)
    cax6 = divider6.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im6, cax=cax6, label=cbarTitle6)
    ax6.set_title(title6)
    ax6.add_artist(ScaleBar(dx=res, units='m'))

    fig.set_size_inches(14, 8)
    fig.suptitle(plotTitle, weight='bold')
    fig.tight_layout(pad=3, w_pad=2.0, h_pad=0.0)
    if savefig == True:
    	plt.show(block=False)
    	figName = plotTitle.replace(' ', '_') + '.png'
    	fig.savefig(figName, dpi=250)
    	# Image.open(figName).show()
    else:
    	plt.show()

def plotDataPoints(dataVals1, cbarTitle1, color1, dataVals2, cbarTitle2, color2,
                   plotTitle, pointx, pointy, pointlabels, savefig=False):
    '''
    Plots a map from input values (array-like) with labelled point locations from reference data
    Plots 2 side by side maps: identical but with difference basemaps (e.g. elevation and pixel mass balance)
    :param dataVals: values to be plotted (array)
    :param cbarTitle: title of the colorbar
    :param color: colorbar scale (https://matplotlib.org/stable/tutorials/colors/colormaps.html)
    :param plotTitle: title of the plot
    :param pointx: x-values of points to plot (raster/array column)
    :param pointy: y-values of points to plot (raster/array row)
    :param pointlabels: text labels for the reference points
    :return:
    '''
    fig = plt.figure()
    ax1 = fig.add_subplot(121, label="1")
    ax2 = fig.add_subplot(122, label="1")

    divnorm1 = TwoSlopeNorm(vmin=dataVals1.min(), vcenter=dataVals1.mean(), vmax=dataVals1.max())
    dataVals1 = dataVals1.astype(float)
    dataVals1[dataVals1 == 0] = np.nan                                  # to remove 0 values for white background in plot
    im1 = ax1.imshow(dataVals1, cmap=plt.cm.get_cmap(color1), norm=divnorm1)  # check len datavals
    divider1 = make_axes_locatable(ax1)
    cax1 = divider1.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im1, cax=cax1, label=cbarTitle1)
    ax1.scatter(pointx, pointy, s=3, c='Red', marker='o')        # add points where reference data exists
    for i, txt in enumerate(pointlabels):                       # add labels to each point
        ax1.annotate(text=txt, xy=(pointx[i][0]+3, pointy[i][0]-3), c='k')

    divnorm2 = TwoSlopeNorm(vmin=dataVals2.min(), vcenter=dataVals2.mean(), vmax=dataVals2.max())
    dataVals2 = dataVals2.astype(float)
    dataVals2[dataVals2 == 0] = np.nan                                  # to remove 0 values for white background in plot
    im2 = ax2.imshow(dataVals2, cmap=plt.cm.get_cmap(color2), norm=divnorm2)  # check len datavals
    divider2 = make_axes_locatable(ax2)
    cax2 = divider2.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im2, cax=cax2, label=cbarTitle2)
    ax2.scatter(pointx, pointy, s=3, c='Red', marker='o')        # add points where reference data exists
    for i, txt in enumerate(pointlabels):                       # add labels to each point
        ax2.annotate(text=txt, xy=(pointx[i][0]+3, pointy[i][0]-3), c='k')

    fig.suptitle(plotTitle, weight='bold')
    fig.tight_layout(pad=3, w_pad=0.0, h_pad=3.0)
    if savefig == True:
    	plt.show(block=False)
    	figName = plotTitle.replace(' ', '_') + '.png'
    	fig.savefig(figName, dpi=250)
    	# Image.open(figName).show()
    else:
    	plt.show()

def velPlot(vx, vy, v_tot, pixel_size, threshold):
    '''
    Show velocity vector arrows
    :param vx1: x-direction velocity
    :param vy1: y-direction velocity
    :param v_tot: magnitude of velocity
    :param pixel_size: pixel size (for density of vectors)
    :param threshold: velocity magnitude threshold for showing arrows
    :return:
    '''
    # fig, ax = plt.subplots()                                  # uncomment lines to plot this graph separately
    if pixel_size < 10:
        freq = 20
    else:
        freq = 10
    xx = np.arange(0, vx.shape[1], freq)                          # last number represents arrow frequency
    yy = np.arange(0, vy.shape[0], freq)
    points = np.ix_(yy, xx)
    px, py = np.meshgrid(xx, yy)

    vx_norm = np.divide(vx[points], v_tot[points], out=np.zeros_like(vx[points]), where=v_tot[points] > threshold)
    vy_norm = np.divide(vy[points], v_tot[points], out=np.zeros_like(vx[points]), where=v_tot[points] > threshold)
    vx_norm[np.isnan(vx_norm)] = 0
    vy_norm[np.isnan(vy_norm)] = 0

    mask = np.logical_or(vx_norm != 0, vy_norm != 0)                # remove 0 points
    quiverInput = [px[mask], py[mask], vx_norm[mask], vy_norm[mask], 1, 'gist_gray', 20]
    # ax.quiver(px[mask], py[mask], vx_norm[mask], vy_norm[mask], 1, cmap='gist_gray', 20)
    #
    # v_tot = v_tot.astype(float)
    # v_tot[v_tot == 0] = np.nan                    # to remove 0 values for white background in plot
    # im = ax.imshow(v_tot, color)
    # fig.colorbar(im, label='Velocity (m/a)')
    # fig.tight_layout(pad=3, w_pad=0, h_pad=0)
    # plt.show(block=False)
    return quiverInput

def elevationBinPlot3data3Subfigs(x_subp1, x_subp1_lab, x_subp1_2, x_subp1_lab_2, x_subp1_3, x_subp1_lab_3, x_subp1_4,
                                  x_subp1_lab_4, y_subp1, x2_subp1, y2_subp1, xLabel1, yLabel1, xLabel2, title1,
                                  elBinWidth, buff, alpha, title, res,
                                  subp3, title3, cbar3, color3,
                                  subp4, title4, cbar4, color4,
                                  subp5, title5, cbar5, color5,
                                  err_subp1_2=None, err_subp1_3=None, err_subp1_4=None,
                                  x_subp1_ref=None, x_subp1_lab_ref=None, y_subp1_ref=None, ref_var=None, cbar4ticks=None, savefig=False):
    '''
    Same as elevationBinPlot3Subfigs (above) but includes comparison of 3 datasets in NW plot
    '''
    fig = plt.figure()
    ax = fig.add_subplot(221, label="1")
    ax2 = fig.add_subplot(261, label="2", frame_on=False)

    ax3 = fig.add_subplot(222)
    ax4 = fig.add_subplot(223)
    ax5 = fig.add_subplot(224)

    ax.scatter(x_subp1, y_subp1, label=x_subp1_lab, color='w', edgecolor='dimgray', zorder=2)
    ax.plot(x_subp1, y_subp1, color='dimgray', zorder=1)
    ax.set_xlabel(xLabel1)
    ax.set_ylabel(yLabel1)
    ax.set_title(title1, pad=10)
    ax.tick_params(axis='x')
    ax.barh(x2_subp1, y2_subp1, elBinWidth, alpha=0, zorder=0)
    if err_subp1_2 is not None:
        if x_subp1_ref is None:
            if len(err_subp1_2.shape) > 1:
                ax.set_xlim([min(np.nanmin(x_subp1_2 - err_subp1_2[0]), np.nanmin(x_subp1_3 - err_subp1_3[0]),
                                 np.nanmin(x_subp1_4 - err_subp1_4[0])) - 50,
                             max(np.nanmax(x_subp1_2 + err_subp1_2[-1]), np.nanmax(x_subp1_3 + err_subp1_3[-1]),
                                 np.nanmax(x_subp1_4 + err_subp1_4[-1])) + 50])
            else:
                ax.set_xlim([min(np.nanmin(x_subp1_2 - err_subp1_2), np.nanmin(x_subp1_3 - err_subp1_3),
                                 np.nanmin(x_subp1_4 - err_subp1_4)) - 50,
                             max(np.nanmax(x_subp1_2 + err_subp1_2), np.nanmax(x_subp1_3 + err_subp1_3),
                                 np.nanmax(x_subp1_4 + err_subp1_4)) + 50])
        else:
            if len(err_subp1_2.shape) > 1:
                ax.set_xlim([min(np.nanmin(x_subp1_2 - err_subp1_2[0]), np.nanmin(x_subp1_3 - err_subp1_3[0]),
                                 np.nanmin(x_subp1_4 - err_subp1_4[0]), np.nanmin(x_subp1_ref)) - 20,
                             max(np.nanmax(x_subp1_2 + err_subp1_2[-1]), np.nanmax(x_subp1_3 + err_subp1_3[-1]),
                                 np.nanmax(x_subp1_4 + err_subp1_4[-1]), np.nanmax(x_subp1_ref), 0) + 20])
            else:
                ax.set_xlim(
                    [min(np.nanmin(x_subp1_2 - err_subp1_2), np.nanmin(x_subp1_3 - err_subp1_3),
                         np.nanmin(x_subp1_4 - err_subp1_4), np.nanmin(x_subp1_ref)) - 20,
                     max(np.nanmax(x_subp1_2 + err_subp1_2), np.nanmax(x_subp1_3 + err_subp1_3),
                         np.nanmax(x_subp1_4 + err_subp1_4), np.nanmax(x_subp1_ref), 0) + 20])
            ax.plot(x_subp1_ref, y_subp1_ref, color='black', zorder=5)  # plot line to connect reference data
    elif x_subp1_ref is not None:
        ax.set_xlim([min(np.min(x_subp1_ref), np.min(x_subp1))-buff, max(np.max(x_subp1_ref), np.max(x_subp1), 0)+buff])
        ax.plot(x_subp1_ref, y_subp1_ref, color='black', zorder=5)  # plot line to connect reference data
    else:
        ax.set_xlim([min(x_subp1) - buff, max(np.max(x_subp1), 0) + buff])

    # ax.set_xlim([-5000, 1500]) # override and just set xlim

    ax2.barh(x2_subp1, y2_subp1, elBinWidth, alpha=alpha, color='dimgray')
    ax2.xaxis.tick_top()
    ax2.set_xlabel(xLabel2, color='dimgray')
    ax2.xaxis.set_label_position('top')
    ax2.tick_params(axis='x', colors='dimgray')

    ax.vlines(x_subp1_2, y_subp1 - np.random.uniform(elBinWidth/2, elBinWidth/2, len(y_subp1)),
              y_subp1 + np.random.uniform(elBinWidth/2, elBinWidth/2, len(y_subp1)), label=x_subp1_lab_2,
              lw=2, color='#01665e', zorder=7)
    ax.errorbar(x_subp1_2, y_subp1 - np.random.uniform(elBinWidth/5, elBinWidth/5, len(y_subp1)), xerr=err_subp1_2,
                fmt='', color='#01665e', ls='none', ecolor='#35978f', elinewidth=2, alpha=0.6, zorder=3)
    ax.vlines(x_subp1_3, y_subp1 - np.random.uniform(elBinWidth/2, elBinWidth/2, len(y_subp1)),
              y_subp1 + np.random.uniform(elBinWidth/2, elBinWidth/2, len(y_subp1)), label=x_subp1_lab_3,
              lw=2, color='#8c510a', zorder=6)
    ax.errorbar(x_subp1_3, y_subp1, xerr=err_subp1_3,
                fmt='', color='#8c510a', ls='none', ecolor='#bf812d', elinewidth=2, alpha=0.6, zorder=4)
    ax.vlines(x_subp1_4, y_subp1 - np.random.uniform(elBinWidth/2, elBinWidth/2, len(y_subp1)),
              y_subp1 + np.random.uniform(elBinWidth/2, elBinWidth/2, len(y_subp1)), label=x_subp1_lab_4,
              lw=2, color='#67000d', zorder=6)
    ax.errorbar(x_subp1_4, y_subp1 + np.random.uniform(elBinWidth/5, elBinWidth/5, len(y_subp1)), xerr=err_subp1_4,
                fmt='', color='#67000d', ls='none', ecolor='#a50f15', elinewidth=2, alpha=0.6, zorder=4)
    ax.scatter(x_subp1_ref, y_subp1_ref, label=x_subp1_lab_ref, color='black', zorder=5)
    ax.errorbar(x_subp1_ref, y_subp1_ref, xerr=ref_var,
                fmt='', color='black', ls='none', ecolor='black', elinewidth=2, alpha=0.6, zorder=3)
    if x_subp1_ref != None:
        m, b = np.polyfit(x_subp1_ref, y_subp1_ref, 1)
        ax.axline((0, b), slope=m, color='black', ls='--', alpha=0.8, zorder=6)
    ax.axvline(x=0, color='black', lw=0.5, dashes=(10, 3))
    ax.legend(fontsize=8)   # bbox_to_anchor=(1.02, 1), loc='upper left', 'best'

    # add 3 subplots
    # top right plot:
    divnorm3 = TwoSlopeNorm(vmin=min(subp3.min(), -1), vcenter=0, vmax=max(subp3.max(), 1))       # center colorbar at 0
    im3 = ax3.imshow(subp3, cmap=plt.cm.get_cmap(color3, len(x_subp1)), norm=divnorm3)   # plot array
    divider3 = make_axes_locatable(ax3)                         # plot on ax3 (northeast subplot)
    cax3 = divider3.append_axes('right', size='5%', pad=0.05)   # specify colorbar axis properties
    fig.colorbar(im3, cax=cax3, label=cbar3)                    # add colorbar
    ax3.set_title(title3, pad=10)                               # add subplot title
    ax3.add_artist(ScaleBar(dx=res, units='m'))

    # bottom left plot:
    divnorm4 = TwoSlopeNorm(vmin=min(subp4.min(), -1), vcenter=0, vmax=max(subp4.max(), 1))  # center colorbar at 0
    # subp4 = subp4.astype(float)
    # subp4[subp4 == 0] = np.nan                                  # to remove 0 values for white background in plot
    im4 = ax4.imshow(subp4, cmap=plt.cm.get_cmap(color4, len(x_subp1)), norm=divnorm4)
    divider4 = make_axes_locatable(ax4)
    cax4 = divider4.append_axes('right', size='5%', pad=0.05)
    cbar4 = fig.colorbar(im4, cax=cax4, label=cbar4)
    if cbar4ticks != None:              # this would be if we want colorbar ticks/labels different from actual values
        cbar4.ax.locator_params(nbins=len(cbar4ticks)//2)
        cbar4.ax.set_yticklabels(np.linspace(cbar4ticks[0], cbar4ticks[-1], len(cbar4ticks)//2).astype(int))
    ax4.set_title(title4, pad=10)
    ax4.add_artist(ScaleBar(dx=res, units='m'))

    # bottom right plot:
    divnorm5 = TwoSlopeNorm(vmin=min(subp5.min(), -1), vcenter=0, vmax=max(subp5.max(), 1))  # center colorbar at 0
    im5 = ax5.imshow(subp5, cmap=plt.cm.get_cmap(color5, len(x_subp1)), norm=divnorm5)
    divider5 = make_axes_locatable(ax5)
    cax5 = divider5.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im5, cax=cax5, label=cbar5)
    ax5.set_title(title5, pad=10)
    ax5.add_artist(ScaleBar(dx=res, units='m'))

    fig.set_size_inches(12, 8)
    fig.tight_layout(pad=0.5, w_pad=-3.0, h_pad=2.0)
    fig.suptitle(title, weight='bold')
    if savefig == True:
    	plt.show(block=False)
    	figName = title.replace(' ', '_') + '.png'
    	fig.savefig(figName, dpi=250)
    	# Image.open(figName).show()
    else:
    	plt.show()

def plotClassify(array, class_vals, class_labels, colors, title, savefig=False):
    # plot array of discrete values
    cmap = ListedColormap(colors)

    # Plot newly classified and masked raster
    fig, ax = plt.subplots(figsize=(10, 5))
    im = ax.imshow(array, cmap=cmap)
    ep.draw_legend(im, titles=class_labels, classes=class_vals)
    ax.set(title=title)
    ax.set_axis_off()

    plt.tight_layout()
    if savefig == True:
    	plt.show(block=False)
    	figName = title.replace(' ', '_') + '.png'
    	fig.savefig(figName, dpi=250)
    	# Image.open(figName).show()
    else:
    	plt.show()

def plotContinuous(array, range, cbar, title, cbarTitle, quiver=None, savefig=False):
    # plot array of continuous values
    fig, ax = plt.subplots(figsize=(10, 5))
    colors = plt.cm.get_cmap(cbar).copy()
    im = ax.imshow(array, cmap=colors, vmin=range[0], vmax=range[1])
    im.cmap.set_under('k')
    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im, cax=cax, label=cbarTitle)
    if quiver != None:  # add velocity arrows if desired
        ax.quiver(quiver[0], quiver[1], quiver[2], quiver[3], quiver[4], cmap=quiver[5], scale=quiver[6], width=.003)

    ax.set(title=title)
    ax.set_axis_off()
    plt.tight_layout()
    if savefig == True:
    	plt.show(block=False)
    	figName = title.replace(' ', '_') + '.png'
    	fig.savefig(figName, dpi=250)
    	# Image.open(figName).show()
    else:
    	plt.show()

def plot_binned_data(binStat, binNumber, outline, remove_top_check=True):
    '''
    Returns data array with values corresponding to alitudinally-aggregated elevation bins
    :param binStat: values of statistic to return in the elevation bin (list)
    :param binNumber: bin number of each value in the array (array-like)
    :param outline: glacier outline (binary; array-like)
    :param remove_top_check: whether to check if the top-most elevation bin is very small, and (if so) removes it. Fixes a potential bug in the altitudinal aggregation (boolean)
    :return: array of altitudinally-aggregated data
    '''
    if remove_top_check==True:
        # If we get a pixel in it's own elevation bin, without a recorded elevation bin. This is a problem
        final_bin_count = np.unique(binNumber, return_counts=True)[1][-1]

        if final_bin_count <= 5 and len(binStat) != int(binNumber.max()):
            print('Top bin pixel count was: ', final_bin_count, '\n\tIt has been merged with the second-to-top bin.')
            binNumber[binNumber == binNumber.max()] = binNumber.max() - 1

    binnedStat = np.zeros(binNumber.shape)

    for i in range(len(binNumber)):
        binnedStat[i] = binStat[binNumber[i] - 1]

    outlineShape = outline.shape
    binnedStat = np.array(binnedStat).reshape(outlineShape) * outline
    return binnedStat




# Glacier functions. Mass balance, area, slope, aspect, hillshade calculations.
# Also some functions correcting velocities
import rasterio
import rasterio.plot
import rasterio.mask
import earthpy.spatial as es
import numpy as np
import math
import matplotlib.pyplot as plt
from scipy import stats
from scipy.ndimage import distance_transform_edt
from shapely.geometry import Point
from raster_basics.SmoothingFunctions import dynamicSmoothing, dynamicSmoothingExponential


def glacierArea(outline, res):
    """
	Area of a glacier
		outline: glacier outline, binary array with 1 denoting on-glacier terrain and 0 denoting off-glacier terrain (array-like)
		res: raster resolution (int or float)
    """
    totalArea = np.sum(outline) * res * res  # area in m2
    return totalArea

def totalMassBalance(dhdt, outline, density):
    """
	Total (dhdt) mass balance (kg/m2-yr) = dhdht (m/yr) * density (kg/m3)
		dhdt: elevation change, m/yr (array-like)
		outline: glacier outline, binary array with 1 denoting on-glacier terrain and 0 denoting off-glacier terrain (array-like)
		density: density assumption, kg/cubic.m (int or float)
    """
    massBalanceTotal = dhdt * density * outline
    return massBalanceTotal

def totalMassBalanceValue(totalMB, area, res):
    """
	Total (dhdt) mass balance value
		totalMB: pixel-wise total mass balance, kg/m2-yr (mm w.e.) (array-like)
		area: glacier area, m2 (int or float)
		res: pixel resolution (int or float)
    """
    # totalMB[totalMB == 0] = np.nan
    totalMassBalanceVal = np.nansum(totalMB * res * res) / area  # total mass balance in kg/m2-yr
    return totalMassBalanceVal

def divQ(vx, vy, h, res, vCol=0.8):
    """
	Flux divergence
		vx, vy: x- and y-direction velocity, m/yr (array-like)
		h: glacier thickness, m (array-like)
		res: pixel resolution (int or float)
		vCol: vertical-average velocity scaling factor, typically 0.8 (int or float)
    """
    divQarray = ((np.gradient(vx, res)[1] * h) + (np.gradient(h, res)[1] * vx) +
                  (np.gradient(vy * -1, res)[0] * h) + (np.gradient(h * -1, res)[0] * vy)) * vCol
    return divQarray
    
def flux_div_comps(vx, vy, h, res, vCol=0.8, filt=True, filter_type='Gauss', filter_factor=4):
    """
    Flux divergence - return each component as well
	vx, vy: x- and y-direction velocity, m/yr (array-like)
	h: glacier thickness, m (array-like)
	res: pixel resolution (int or float)
	vCol: vertical-average velocity scaling factor, typically 0.8 (int or float)
	filt: whether or not to smooth derivatives
	filter_type: type of filter, either 'Gauss' or 'Exp'
	filter_factor: smoothing factor for filtering
    """
    dvx = (np.gradient(vx, res)[1] * h) * vCol
    dvy = (np.gradient(vy * -1, res)[0] * h) * vCol
    dhx = (np.gradient(h, res)[1] * vx) * vCol
    dhy = (np.gradient(h * -1, res)[0] * vy) * vCol
    if filt == True:
        if filter_type == 'Gauss':
            dvx = dynamicSmoothing(dvx, h, res, filter_factor)
            dvy = dynamicSmoothing(dvy, h, res, filter_factor)
            dhx = dynamicSmoothing(dhx, h, res, filter_factor)
            dhy = dynamicSmoothing(dhy, h, res, filter_factor)
        elif filter_type == 'Exp':
            dvx = dynamicSmoothingExponential(dvx, h, res, filter_factor)
            dvy = dynamicSmoothingExponential(dvy, h, res, filter_factor)
            dhx = dynamicSmoothingExponential(dhx, h, res, filter_factor)
            dhy = dynamicSmoothingExponential(dhy, h, res, filter_factor)
    divQ = dvx + dvy + dhx + dhy
    return divQ, dvx, dvy, dhx, dhy


# import richdem as rd
# def glacierAttributes(dem_rast, attrFormat):
#    # return desired attribute (attrFormat is 'slope_degree' or 'slope_percentage', 'slope_riserun', 'aspect')
#    # https://richdem.readthedocs.io/en/latest/python_api.html#richdem.rdarray
#    # gdal.DEMProcessing('slope.tif', dem_rast, 'slope', slopeFormat=slopeFormat)
#    # with rasterio.open('slope.tif') as dataset:
#    #     slope = dataset.read(1)
#    #     slope[slope == -9999.0] = 0
#    no_data_val = rasterio.open(dem_rast).nodatavals[0]
#    if no_data_val == None:
#        dem_array = rd.LoadGDAL(dem_rast, no_data=-9999.0)  # assign a No Data value if none is prescribed
#    else:
#        dem_array = rd.LoadGDAL(dem_rast)
#    attr_array = rd.TerrainAttribute(dem_array, attrib=attrFormat)
#    return attr_array

def glacierSlope(array, res):
    """
	Rise/run slope of a raster array
		array: input array (array-like)
		res: pixel resolution (int or float)
    """
    # alternate way to obtain the riserun slope, with an array input instead of raster
    px, py = np.gradient(array, res)
    slope = np.sqrt(px ** 2 + py ** 2)
    return slope

def demHillshade(demArray, az=135, alt=30):
    # https://earthpy.readthedocs.io/en/latest/gallery_vignettes/plot_dem_hillshade.html
    # demArray is an input raster DEM opened and read as an array
    # azimuth 90 is east, 0 is north. default value 135 is SE for Alaska
    # altitude is from 0 to 90, 90 being directly overhead
    hillshade = es.hillshade(demArray, azimuth=az, altitude=alt)     # Create and plot the hillshade with earthpy
    return hillshade
    
def demAspect(dem_array, res):
    """
    Aspect of a DEM array – takes an array input instead of raster
	dem_array: input array (array-like)
	res: pixel resolution (int or float)
    """
    aspect_array = np.zeros_like(dem_array)
    py, px = np.gradient(dem_array, res)

    for i in range(len(aspect_array)):
        for j in range(len(aspect_array[0])):
            pixel_deg = math.atan2(py[i][j], px[i][j]) * 180 / math.pi
            if pixel_deg >= 0 and pixel_deg <= 90:
                pixel_aspect = 90 - pixel_deg
            elif pixel_deg < 0:
                pixel_aspect = abs(pixel_deg) + 90
            elif pixel_deg > 90:
                pixel_aspect = 450 - pixel_deg
            aspect_array[i][j] = pixel_aspect
    return aspect_array


def velFlowlineAspect(linestring, array, arr_extent, mask=None):
    '''
    Use flowline as velocity direction. Returns array of aspects, based on the flowline
    # linestring: center flowline json object. Type: shapely.geometry.linestring.LineString
    # array: 2D numpy array, output aspect will be the same size
    # arr_extent: Array extent in shapely geometry coordinates (left, bottom, right, top) 
    # mask: Glacier outline mask, 2D array (default None considers all values in arry)
    # returns: 2D array of velocity aspect based on flowline
    '''
    aspect_array = np.zeros_like(array)
    x, y = np.meshgrid(np.arange(arr_extent[0], arr_extent[2]+1), np.arange(arr_extent[1], arr_extent[3]+1))
    if mask is None:
        mask = np.ones_like(array)
    
    # iterate through velocity array
    for i in range(len(array)): 
        for j in range(len(array[0])):
            if mask[i][j] == 1:
                point = Point(x[i,j], y[i,j]) # find our point in json coordinates
                closest_point = linestring.interpolate(linestring.project(point)) # find the nearest point on linestring
                coords = np.array(linestring.coords)
                dists = np.linalg.norm(coords - closest_point.coords, axis=1)
                closest_idx = np.argmin(dists) # find the index of the closest point on the line

                if closest_idx == 0: # check if nearest point is endpoint
                    prev_point = coords[closest_idx]
                    next_point = coords[closest_idx + 2]
                elif closest_idx == len(np.array(linestring.coords)) - 1:
                    prev_point = coords[closest_idx - 2]
                    next_point = coords[closest_idx]
                else:
                    prev_point = coords[closest_idx - 1]
                    next_point = coords[closest_idx + 1]

                # angle of perpendicular line -- perpendicular line is -1/m, so we input (-x, y) instead of (y, x)
                    # perpendicular line is important because we shift aspect from normal coordinates (east is 0)
                    # to DEM aspect (north is 0)
                aspect_angle = np.rad2deg(np.arctan2(prev_point[0] - next_point[0], next_point[1] - prev_point[1]))
                aspect_angle %= 360 # adjust range from 0 to 360
                aspect_array[i][j] = aspect_angle # add to array of aspects
    return aspect_array


def weighted_average(value1, value2, weight1, weight2):
    # Convert directions to radians
    value1_rad = np.deg2rad(value1)
    value2_rad = np.deg2rad(value2)

    # Convert directions to complex numbers
    value1_complex = np.exp(1j * value1_rad)
    value2_complex = np.exp(1j * value2_rad)

    # Calculate the weighted average of complex numbers
    average_complex = (weight1 * value1_complex + weight2 * value2_complex) / (weight1 + weight2)

    # Convert the average complex number back to direction in radians
    average_rad = np.angle(average_complex)

    # Convert the average direction to degrees
    average_deg = np.rad2deg(average_rad)

    # Wrap the average direction back to the range of -180 to 180
    if average_deg > 180:
        average_deg -= 360

    return average_deg


def velFlowlineOutlineAspect(linestring, outlinestring, relative_distance_array, arr_extent, mask):
    '''
    Use flowline/outline as velocity direction. Returns array of aspects, based on the flowline and outline.
    The direction is a combination of the flowline/outline direction, weighted by the distance from each
    # linestring: center flowline json object. Type: shapely.geometry.linestring.LineString
    # outlinestring: glacier outline json object. Type: shapely.geometry.linestring.LineString
    # relative_distance_array: 2D numpy array, array of relative distance from centerline to outline (0 to 1)
    # arr_extent: Array extent in shapely geometry coordinates (left, bottom, right, top) 
    # mask: Glacier outline mask, 2D array (default None considers all values in arry)
    # returns: 2D array of velocity aspect based on flowline and outline
    '''
    aspect_array = np.zeros_like(relative_distance_array)
    
    x, y = np.meshgrid(np.arange(arr_extent[0], arr_extent[2]+1), np.arange(arr_extent[1], arr_extent[3]+1))
    if mask is None:
        mask = np.ones_like(relative_distance_array)
    
    # iterate through velocity array
    for i in range(len(relative_distance_array)): 
        for j in range(len(relative_distance_array[0])):
            if mask[i][j] == 1:
                point = Point(x[i,j], y[i,j]) # find our point in json coordinates

                # # aspect of the centerline
                closest_point_cl = linestring.interpolate(linestring.project(point)) # find the nearest point on line
                coords_cl = np.array(linestring.coords)
                dists_cl = np.linalg.norm(coords_cl - closest_point_cl.coords, axis=1)
                closest_idx_cl = np.argmin(dists_cl) # find the index of the closest point on the line

                if closest_idx_cl == 0: # check if nearest point is endpoint
                    prev_point_cl = coords_cl[closest_idx_cl]
                    next_point_cl = coords_cl[closest_idx_cl + 2]
                elif closest_idx_cl == len(np.array(linestring.coords)) - 1:
                    prev_point_cl = coords_cl[closest_idx_cl - 2]
                    next_point_cl = coords_cl[closest_idx_cl]
                else:
                    prev_point_cl = coords_cl[closest_idx_cl - 1]
                    next_point_cl = coords_cl[closest_idx_cl + 1]

                # angle of perpendicular line -- perpendicular line is -1/m, so we input (-x, y) instead of (y, x)
                aspect_angle_cl = np.rad2deg(np.arctan2(prev_point_cl[0] - next_point_cl[0], 
                                                        next_point_cl[1] - prev_point_cl[1]))
                aspect_angle_cl %= 360 # adjust range from 0 to 360

                # # aspect of the outline
                closest_point_ol = outlinestring.interpolate(outlinestring.project(point))
                coords_ol = np.array(outlinestring.coords)
                dists_ol = np.linalg.norm(coords_ol - closest_point_ol.coords, axis=1)
                closest_idx_ol = np.argmin(dists_ol)
                
                if closest_idx_ol == len(np.array(outlinestring.coords)) - 1:
                    prev_point_ol = coords_ol[closest_idx_ol - 1]
                    next_point_ol = coords_ol[0]
                else:
                    prev_point_ol = coords_ol[closest_idx_ol - 1]
                    next_point_ol = coords_ol[closest_idx_ol + 1]
                # angle of perpendicular line -- perpendicular line is -1/m, so we input (-x, y) instead of (y, x)
                aspect_angle_olr = np.arctan2(prev_point_ol[0] - next_point_ol[0],
                                              next_point_ol[1] - prev_point_ol[1])

                # get aspect of the angle and adjust range from 0 to 360
                aspect_angle_ol = np.rad2deg(aspect_angle_olr)
                aspect_angle_ol %= 360
                
                # make sure aspect is in the right direction (the angle should be closer to the centerline angle)
                if abs(aspect_angle_ol - aspect_angle_cl + 180) < abs(aspect_angle_ol - aspect_angle_cl):
                    aspect_angle_ol += 180
                elif abs(aspect_angle_ol - aspect_angle_cl - 180) < abs(aspect_angle_ol - aspect_angle_cl):
                    aspect_angle_ol -= 180
                aspect_angle_ol %= 360

                # weighted average based on relative distance  
                aspect_array[i][j] = weighted_average(aspect_angle_ol, aspect_angle_cl, relative_distance_array[i][j], 
                                                      (1 - relative_distance_array[i][j]))
    return aspect_array


def velocityAspect(vel_x, vel_y):
    # gets the aspect of velocity vectors, given input x- and y-direction velocity arrays
    vel_aspect = np.zeros_like(vel_x)
    for i in range(len(vel_aspect)):
        for j in range(len(vel_aspect[0])):
            pixel_deg = math.atan2(vel_y[i][j], vel_x[i][j]) * 180 / math.pi
            if pixel_deg >= 0 and pixel_deg <= 90:
                pixel_aspect = 90 - pixel_deg
            elif pixel_deg < 0:
                pixel_aspect = abs(pixel_deg) + 90
            elif pixel_deg > 90:
                pixel_aspect = 450 - pixel_deg
            elif math.isnan(pixel_deg) == True:
                pixel_aspect = math.nan
            vel_aspect[i][j] = pixel_aspect
    return vel_aspect

def velocityAspectAngle(vel_aspect1, vel_aspect2):
    # gets the angle between an aspect array and an ARRAY of aspect arrays. returns the maximum
    for i in range(len(vel_aspect2)):
        if i == 1:
            b = a.copy()
        a = np.arccos(np.cos((vel_aspect1 - vel_aspect2[i]) * math.pi / 180)) * 180 / math.pi
        if i > 0:
            b = np.maximum(a, b)
    return b


def velAspectCorrection(dem_aspect, vel_x, vel_y, threshold=90):
    # replace velocities that go against the aspect ('uphill') with 0
    # first, find the aspect based on the velocity (save this in vel_aspect)
    vel_aspect = velocityAspect(vel_x, vel_y)

    # now, compare and replace values where aspect differs beyond the threshold
    vel_x_cor = np.zeros_like(vel_x)
    vel_y_cor = np.zeros_like(vel_y)
    for i in range(len(dem_aspect)):
        for j in range(len(dem_aspect[0])):
            dem_aspect_high = dem_aspect[i][j] + 360
            dem_aspect_low = dem_aspect[i][j] - 360
            # find the locations where we are within the threshold
            if np.array([abs(dem_aspect[i][j] - vel_aspect[i][j]) <= threshold,
                         abs(dem_aspect_high - vel_aspect[i][j]) <= threshold,
                         abs(dem_aspect_low - vel_aspect[i][j]) <= threshold]).any():
                vel_x_cor[i][j] = vel_x[i][j]
                vel_y_cor[i][j] = vel_y[i][j]
            else:       # if we are beyond the threshold, we return 0
                # vel_x_cor[i][j] = vel_x[i][j] * 0.000001
                # vel_y_cor[i][j] = vel_y[i][j] * 0.000001
                vel_x_cor[i][j] = vel_x[i][j] * 0.0000001
                vel_y_cor[i][j] = vel_y[i][j] * 0.0000001
    return vel_x_cor, vel_y_cor

def velAspectDirection(dem_aspect, vel):
    # use velocity magnitudes, but use aspect for the direction
    vel_x_cor = np.zeros_like(vel)
    vel_y_cor = np.zeros_like(vel)
    for i in range(len(dem_aspect)):
        for j in range(len(dem_aspect[0])):
            # convert from aspect and vel magnitude to vx and vy
            vel_x_cor[i][j] = np.sin(dem_aspect[i][j] * math.pi / 180) * vel[i][j]  # this is our new vel_x
            vel_y_cor[i][j] = np.cos(dem_aspect[i][j] * math.pi / 180) * vel[i][j]  # this is our new vel_y
    return vel_x_cor, vel_y_cor

def velAspectSlopeThreshold(dem_aspect, vel_x, vel_y, dem_slope, slope_threshold):
    # use velocity magnitudes, but use aspect for the direction IF SLOPE EXCEEDS A THRESHOLD
    # compare and replace values where aspect differs beyond the threshold
    vel_x_cor = np.zeros_like(vel_x)
    vel_y_cor = np.zeros_like(vel_y)
    vel = np.power((np.square(vel_x) + np.square(vel_y)), 0.5)
    for i in range(len(dem_aspect)):
        for j in range(len(dem_aspect[0])):
            # convert from aspect and vel magnitude to vx and vy
            if dem_slope[i][j] > slope_threshold:
                vel_x_cor[i][j] = np.sin(dem_aspect[i][j] * math.pi / 180) * vel[i][j]  # this is our new vel_x
                vel_y_cor[i][j] = np.cos(dem_aspect[i][j] * math.pi / 180) * vel[i][j]  # this is our new vel_y
            else:
                vel_x_cor[i][j] = vel_x[i][j]
                vel_y_cor[i][j] = vel_y[i][j]
    return vel_x_cor, vel_y_cor

def velAspectSlopeAverage(dem_aspect, vel_x, vel_y, dem_slope, slope_weight):
    # use velocity magnitudes, but calculate direction as an average of aspect and raw data weighted based on slope
    # slope_weight is the slope value where raw data and DEM-based aspect have even weight. higher values attribute
    #   more weight to the raw data based aspect
    # compare and replace values where aspect differs beyond the threshold
    vel_x_cor = np.zeros_like(vel_x)
    vel_y_cor = np.zeros_like(vel_y)
    vel = np.power((np.square(vel_x) + np.square(vel_y)), 0.5)
    for i in range(len(dem_aspect)):
        for j in range(len(dem_aspect[0])):
            # first, find the aspect based on the velocity (save this in vel_aspect)
            pixel_deg = math.atan2(vel_y[i][j], vel_x[i][j]) * 180 / math.pi
            if pixel_deg >= 0 and pixel_deg <= 90:
                pixel_aspect = 90 - pixel_deg
            elif pixel_deg < 0:
                pixel_aspect = abs(pixel_deg) + 90
            elif pixel_deg > 90:
                pixel_aspect = 450 - pixel_deg
            # wieghted aspect of velocity using DEM and raw velocity product direction
            weighted_aspect = np.average([pixel_aspect, dem_aspect[i][j]], weights=[1, dem_slope[i][j]/slope_weight])
            vel_x_cor[i][j] = np.sin(weighted_aspect * math.pi / 180) * vel[i][j]  # this is our new vel_x
            vel_y_cor[i][j] = np.cos(weighted_aspect * math.pi / 180) * vel[i][j]  # this is our new vel_y
    return vel_x_cor, vel_y_cor

def slope_vel_plot(dem_slope, vel, title, slope_threshold=60, showPlot=False):
    # plot the difference in velocity magnitudes vs slope
    # filter out slope values larger than 60, where ice can't stick to, or a input threshold
    dem_slope_masked = np.ma.masked_where(dem_slope > min(slope_threshold, 60), dem_slope)
    vel_masked = np.ma.masked_where(np.ma.getmask(dem_slope_masked), vel)  # apply the mask to the velocity data too
    new_vel = vel_masked / np.cos(dem_slope_masked * math.pi / 180)  # correction for map-velocity vs in-plane velocity

    fig, ax = plt.subplots()
    ax.scatter(dem_slope_masked.flatten(), new_vel.flatten(), s=1, c='r', alpha=0.25)

    m, b = np.polyfit(dem_slope_masked.flatten(), new_vel.flatten(), 1)
    plt.plot(dem_slope_masked.flatten(), m * dem_slope_masked.flatten() + b, color='k', lw=0.5)

    mean, boundary, number = stats.binned_statistic(dem_slope_masked.flatten(), new_vel.flatten(),
                                                    statistic='mean', bins=np.linspace(0, 60, 24))
    ax.hlines(mean, boundary[1:], boundary[:-1], colors='b', alpha=1)

    ax.set_xlabel('Slope (degrees)')
    ax.set_ylabel('Velocity Magnitude (In-Plane) (m/a)')
    ax.set_title(title, weight='bold')
    ax.set_xlim(left=0, right=dem_slope_masked.max())
    ax.set_ylim(bottom=0, top=new_vel.max())
    plt.grid()
    plt.show(block=showPlot)

def h_vel_plot(dem_slope, thickness, vel, title, slope_threshold=60, showPlot=False):
    # plot the difference in velocity magnitudes vs slope
    # filter out slope values larger than 60, where ice can't stick to, or a input threshold
    dem_slope_masked = np.ma.masked_where(dem_slope > min(slope_threshold, 60), dem_slope)
    vel_masked = np.ma.masked_where(np.ma.getmask(dem_slope_masked), vel)  # apply the mask to the velocity data too
    thickness_masked = np.ma.masked_where(np.ma.getmask(dem_slope_masked), thickness)
    new_vel = vel_masked / np.cos(dem_slope_masked * math.pi / 180)  # correction for map-velocity vs in-plane velocity

    fig, ax = plt.subplots()
    ax.scatter(thickness_masked.flatten(), new_vel.flatten(), s=1, c='r', alpha=0.25)

    m, b = np.polyfit(thickness_masked.flatten(), new_vel.flatten(), 1)
    plt.plot(thickness_masked.flatten(), m * thickness_masked.flatten() + b, color='k', lw=0.5)

    mean, boundary, number = stats.binned_statistic(thickness_masked.flatten(), new_vel.flatten(),
                                                    statistic='mean', bins=np.linspace(0, 220, 21))
    ax.hlines(mean, boundary[1:], boundary[:-1], colors='b', alpha=1)

    ax.set_xlabel('Thickness (m)')
    ax.set_ylabel('Velocity Magnitude (In-Plane) (m/a)')
    ax.set_title(title, weight='bold')
    ax.set_xlim(left=0, right=thickness_masked.max())
    ax.set_ylim(bottom=0, top=new_vel.max())
    plt.grid()
    plt.show(block=showPlot)

def stress_vel_plot(dem_slope, thickness, vel, title, slope_threshold=60, showPlot=False):
    # plot the difference in velocity magnitudes vs driving stress (slope * thickness)
    # filter out slope values larger than 60, where ice can't stick to, or a input threshold
    dem_slope_masked = np.ma.masked_where(dem_slope > min(slope_threshold, 60), dem_slope)
    vel_masked = np.ma.masked_where(np.ma.getmask(dem_slope_masked), vel)  # apply the mask to the velocity data too
    thickness_masked = np.ma.masked_where(np.ma.getmask(dem_slope_masked), thickness)
    driving_stress = dem_slope_masked * thickness_masked
    new_vel = vel_masked / np.cos(dem_slope_masked * math.pi / 180)  # correction for map-velocity vs in-plane velocity

    fig, ax = plt.subplots()
    ax.scatter(driving_stress.flatten(), new_vel.flatten(), s=1, c='r', alpha=0.25)

    mean, boundary, number = stats.binned_statistic(driving_stress.flatten(), new_vel.flatten(),
                                                    statistic='mean', bins=np.linspace(0, 4000, 20))
    ax.hlines(mean, boundary[1:], boundary[:-1], colors='b', alpha=1)

    m, b = np.polyfit(driving_stress.flatten(), new_vel.flatten(), 1)
    plt.plot(driving_stress.flatten(), m * driving_stress.flatten() + b, color='k', lw=0.5)

    ax.set_xlabel('Driving Stress (m-deg)')
    ax.set_ylabel('Velocity Magnitude (In-Plane) (m/a)')
    ax.set_title(title, weight='bold')
    ax.set_xlim(left=0, right=driving_stress.max())
    ax.set_ylim(bottom=0, top=new_vel.max())
    plt.grid()
    plt.show(block=showPlot)


def distance_from_line(line_array, res, mask=None):
    '''
    Calculate the Euclidean distance from centerline array
    :param line_array: Boolean array input to calculate distance from. Distance is calculate from true values
    :param res: Array Euclidean spacing (resoluion) -- numerical
    :param mask: Boolean array mask for on/off glacier terrain (default None takes entire array as on-glacier)
    :return: Array of distance values from line array  
    '''
    
    if mask is None: # if not mask is input, we create a mask that includes all values in array
        mask = np.ones_like(line_array) 
        
    distance_array = distance_transform_edt(np.logical_not(line_array), sampling=[res, res]) # calculate distance
    distance_array[mask == 0] = 0 # remove off-glacier terrain
    return distance_array




from raster_basics import RasterBasics
from raster_basics import DataPlots
from raster_basics import GlacierFunctions
from raster_basics import BaseFunctions
from raster_basics import SmoothingFunctions

# import raster_basics.show_fig
# import raster_basics.shpReprojection
# import raster_basics.shpClip
# import raster_basics.tifReprojectionResample
# import raster_basics.fillHole
# import raster_basics.mosaic_files
# import raster_basics.rasterLike
# import raster_basics.extract_along_line
# import raster_basics.points_along_lines
# import raster_basics.end_points




import numpy as np


def add_grf_noise(array, array_res, correlation_length, noise_strength):
    '''
    Returns the array with spatially-correlated noise added from a Gaussian Random Field
    :param array: np array that we want to add noise to (2D np.array)
    :param array_res: array resolution (numeric, represents meters)
    :param correlation_length: length of spatial correlation in noise (numeric, represents meters)
    :param noise_strength: magnitude of noise. This is multiplied by a GRF with zero mean and unit variance (numeric)
    '''
    mask = np.isnan(array) # off-glacier mask (nan values)
    correlation_length_eff = correlation_length / array_res # effective correlation length (pixel length)
    
    # --- Generate the Gaussian Random Field ---
    distances = np.meshgrid(*[np.arange(dim) for dim in array.shape], indexing='ij') # Generate a grid of distances
    squared_distances = sum(dist**2 for dist in distances)

    covariance = np.exp(-0.5 * squared_distances / correlation_length_eff**2) # Compute the covariance function
    random_field = np.random.normal(size=array.shape) # Generate a random field with zero mean and unit variance
    fft_field = np.fft.fftn(random_field) # Perform a Fast Fourier Transform
    filtered_fft_field = np.fft.fftn(covariance**0.5) * fft_field # sq.rt of the covariance function in Fourier space
    grf_noise = np.fft.ifftn(filtered_fft_field).real # Perform an inverse Fast Fourier Transform for GRF
    # --- --- --- --- --- --- --- --- --- --- ---
    
    grf_noise_norm = (grf_noise - np.mean(grf_noise)) / np.std(grf_noise) # normalize GRF (zero mean, unit variance)

    scaled_grf = noise_strength * grf_noise_norm # Scale the noise field by the desired noise strength
    grf_array_noise = np.where(mask, array, array + scaled_grf) # Add the noise to vel array
    array_noisy = np.where(grf_array_noise < 0, 0, grf_array_noise) # Ensure that velocity values remain positive
    return array_noisy


def add_ice_thickness_bias(ice_thickness, relative_distance, comp_factor=2, direction='UV'):
    """
    Add bias to the ice thickness data based on the relative distance from the glacier center. This assumes bias in ice thickness 
    relative to distance from centerline resulting from inaccurate bed profile assumption during ice thickness inversion

    :param ice_thickness: array of ice thickness values. (np array)
    :param relative_distance: array of relative distances from the glacier centerline to edge (np array)
    :param comp_factor: comparison factor for the V to U-shaped cross-sections. Default of 2 gives max of ~30% thickness change. Lower values add greater bias (numeric)
    :param direction: direction of bias, either 'UV' or 'VU'. 'UV' assumes our modeled ice thickness is a parabola and reality is a V. 'VU' assumes the opposite case (str)
    
    Returns: array of ice thickness values with added bias, and the bias factor
    """

    # apply bias based on the relative distance 
    # we use a 2nd order parabolic function, following from the shallow-ice approximation (consistent with OGGM)
    # we use a V-shape function with the same mean as the parabola
    bias_factor = ((relative_distance ** 2) + (1/3) + comp_factor) / ((4/3) * relative_distance + comp_factor)

    # Add the bias to the ice thickness data
    if direction == 'VU':
        bias_factor = 1 / bias_factor
    
    ice_thickness_bias = ice_thickness * bias_factor

    return ice_thickness_bias, bias_factor





import numpy as np
import scipy.signal
import scipy.ndimage
from scipy.stats import binned_statistic
import itertools
import rasterio
from .RasterBasics import points_along_lines

def sgolay2d(z, window_size, order, derivative=None):
    '''
    Savitzky-Golay smoothing filter for eliminating high frequency noise from data via moving average technique
    Type of low-pass filter
    :param z: input file array
    :param window_size: odd number: window size of smoothing
    :param order: order of filtering polynomial (cannot be greater than window_size squared)
    :param derivative: calculate the derivative as well: gradient. entries can be 'col', 'row', or 'both'
    :return: smoothed signal
    '''
    # number of terms in the polynomial expression
    n_terms = (order + 1) * (order + 2) / 2.0

    if window_size % 2 == 0:
        raise ValueError('window_size must be odd')
    if window_size ** 2 < n_terms:
        raise ValueError('order is too high for the window size')

    half_size = window_size // 2

    # exponents of the polynomial.
    # p(x,y) = a0 + a1*x + a2*y + a3*x^2 + a4*y^2 + a5*x*y + ...
    # this line gives a list of two item tuple. Each tuple contains
    # the exponents of the k-th term. First element of tuple is for x
    # second element for y.
    # Ex. exps = [(0,0), (1,0), (0,1), (2,0), (1,1), (0,2), ...]
    exps = [(k - n, n) for k in range(order + 1) for n in range(k + 1)]

    # coordinates of points
    ind = np.arange(-half_size, half_size + 1, dtype=np.float64)
    dx = np.repeat(ind, window_size)
    dy = np.tile(ind, [window_size, 1]).reshape(window_size ** 2, )

    # build matrix of system of equation
    A = np.empty((window_size ** 2, len(exps)))

    for i, exp in enumerate(exps):
        A[:, i] = (dx ** exp[0]) * (dy ** exp[1])

    # pad input array with appropriate values at the four borders
    new_shape = z.shape[0] + 2 * half_size, z.shape[1] + 2 * half_size
    Z = np.zeros((new_shape))
    # top band
    band = z[0, :]
    Z[:half_size, half_size:-half_size] = band - np.abs(np.flipud(z[1:half_size + 1, :]) - band)
    # bottom band
    band = z[-1, :]
    Z[-half_size:, half_size:-half_size] = band + np.abs(np.flipud(z[-half_size - 1:-1, :]) - band)
    # left band
    band = np.tile(z[:, 0].reshape(-1, 1), [1, half_size])
    Z[half_size:-half_size, :half_size] = band - np.abs(np.fliplr(z[:, 1:half_size + 1]) - band)
    # right band
    band = np.tile(z[:, -1].reshape(-1, 1), [1, half_size])
    Z[half_size:-half_size, -half_size:] = band + np.abs(np.fliplr(z[:, -half_size - 1:-1]) - band)
    # central band
    Z[half_size:-half_size, half_size:-half_size] = z
    # top left corner
    band = z[0, 0]
    Z[:half_size, :half_size] = band - np.abs(np.flipud(np.fliplr(z[1:half_size + 1, 1:half_size + 1])) - band)
    # bottom right corner
    band = z[-1, -1]
    Z[-half_size:, -half_size:] = band + np.abs(np.flipud(np.fliplr(z[-half_size - 1:-1, -half_size - 1:-1])) - band)
    # top right corner
    band = Z[half_size, -half_size:]
    Z[:half_size, -half_size:] = band - np.abs(np.flipud(Z[half_size + 1:2 * half_size + 1, -half_size:]) - band)
    # bottom left corner
    band = Z[-half_size:, half_size].reshape(-1, 1)
    Z[-half_size:, :half_size] = band - np.abs(np.fliplr(Z[-half_size:, half_size + 1:2 * half_size + 1]) - band)

    # solve system and convolve
    if derivative == None:
        m = np.linalg.pinv(A)[0].reshape((window_size, -1))
        return scipy.signal.fftconvolve(Z, m, mode='valid')
    elif derivative == 'col':
        c = np.linalg.pinv(A)[1].reshape((window_size, -1))
        return scipy.signal.fftconvolve(Z, -c, mode='valid')
    elif derivative == 'row':
        r = np.linalg.pinv(A)[2].reshape((window_size, -1))
        return scipy.signal.fftconvolve(Z, -r, mode='valid')
    elif derivative == 'both':
        c = np.linalg.pinv(A)[1].reshape((window_size, -1))
        r = np.linalg.pinv(A)[2].reshape((window_size, -1))
        return scipy.signal.fftconvolve(Z, -r, mode='valid'), scipy.signal.fftconvolve(Z, -c, mode='valid')

def gaussianFilter(z, st_dev, truncate=4, fill_mode='constant', fill_val=0):
    '''
    Gaussian smoothing filter
    :param z: input file (open, read raster file; array-like)
    :param st_dev: standard deviation for filter. higher value indicates greater blur
    :param fill_mode: parameter that determines how the input array is extended when the filter overlaps a border.
    :param fill_val: value to fill past edges of input if mode is ‘constant’. Default is 0.0.
    :return: new, smoothed array
    '''
    pixel_w = 2 * int(truncate * st_dev + 0.5) + 1      # compute pixel window width
    #print('Gaussian Filter Window Width is', pixel_w)
    gaussian = scipy.ndimage.filters.gaussian_filter(z, st_dev, truncate=truncate, mode=fill_mode, cval=fill_val)
    return gaussian

def dynamicSmoothing(z, window_weight, pixel_size, f):
    '''
    Gaussian filter, dynamic smoothing with changing pixel window size
    :param z: input file (array)
    :param window_weight: weight of window size (thickness file)
    :param pixel_size: size of pixel (resolution)
    :param f: factor of window size from window weight (4 for vx, vy, h. 1 for divQ)
    :return: array of smoothed values
    '''

    # Apply a border of zeros based on the half size
    half_size_max = int(np.max(window_weight) / pixel_size) * f
    pad = np.max(half_size_max) + 2
    r = z.shape[0]
    c = z.shape[1]
    lr_border = np.zeros((r, pad), dtype=int)                    # left/right border
    tb_border = np.zeros((pad, c + (2*pad)), dtype=int)          # top/bottom border
    x = z.copy()
    x = np.concatenate((lr_border, x), axis=1)
    x = np.concatenate((x, lr_border), axis=1)
    x = np.concatenate((tb_border, x), axis=0)
    x = np.concatenate((x, tb_border), axis=0)

    smooth_file = np.zeros(z.shape)         # initiate array to store each smooth pixel
    for r in range(z.shape[0]):
        for c in range(z.shape[1]):
            if window_weight[r, c] != 0:
                # half of window size (must be at least f * pixels)
                half_size = int(f * max(window_weight[r, c], pixel_size) / pixel_size)
                # add the border size to start at actual values
                pixel_window = x[r+pad-half_size:r+pad+half_size+1, c+pad-half_size:c+pad+half_size+1]

                # Gaussian Filter Params
                truncate = 3
                st_dev = (half_size - 0.5) / truncate          # filter st.dev depends on desired window size at a pixel
                smooth_array = scipy.ndimage.filters.gaussian_filter(pixel_window, st_dev, truncate=truncate)
                smooth_file[r, c] = smooth_array[half_size, half_size]      # select the center value as the smoothed value
            else:
                smooth_file[r, c] = 0           # we don't want to smooth 0 regions so we leave these as 0
    return smooth_file

def dynamicSmoothingExponential(z, window_weight, pixel_size, f, max_size=2500):
    '''
    Exponential filter, dynamic smoothing with changing pixel window size
    :param z: input file (array)
    :param window_weight: weight of window size (thickness file)
    :param pixel_size: size of pixel (resolution)
    :param f: factor of window size from window weight (4 for vx, vy, h. 1 for divQ)
    :param max_size: maximum window size for smoothing (int)
    :return: array of smoothed values
    '''

    # Apply a border of zeros based on the half size
    half_size = int((max_size / pixel_size) + 1)
    pad = np.max(half_size) + 2
    r = z.shape[0]
    c = z.shape[1]
    lr_border = np.zeros((r, pad), dtype=int)                    # left/right border
    tb_border = np.zeros((pad, c + (2*pad)), dtype=int)          # top/bottom border
    x = z.copy()
    x = np.concatenate((lr_border, x), axis=1)
    x = np.concatenate((x, lr_border), axis=1)
    x = np.concatenate((tb_border, x), axis=0)
    x = np.concatenate((x, tb_border), axis=0)
    
    row_i, col_j = np.indices((half_size*2+1, half_size*2+1))
    smooth_file = np.zeros(z.shape)         # initiate array to store each smooth pixel
    for r in range(z.shape[0]):
        for c in range(z.shape[1]):
            if window_weight[r, c] != 0:
                # add the border size to start at actual values
                pixel_window = x[r+pad-half_size:r+pad+half_size+1, c+pad-half_size:c+pad+half_size+1]
                exp_weights = np.exp(-np.sqrt(((half_size - row_i)*pixel_size)**2 + 
                                              ((half_size - col_j)*pixel_size)**2)/(f*window_weight[r,c]))
                exp_weights /= np.sum(exp_weights)  # adjust weights to sum to 1
                smooth_file[r, c] = np.sum(pixel_window*exp_weights)
            else:
                smooth_file[r, c] = 0           # we don't want to smooth 0 regions so we leave these as 0
    return smooth_file

def smoothingCorrection(orig_data, smooth_data, centerlines):
    '''
    Apply a correction to smoothed data products, such that smoothing does not decrease the mean values
    Correction is based on glacier centerline values
    :param orig_data: originial data, e.g. unsmoothed velocity or ice thickness (raster)
    :param smooth_data: smoothed data, e.g. output after applying Gaussian filter (raster)
    :param centerlines: centerline(s) for the data, e.g. from OGGM (shapefile)
    :return: smooth array that has been corrected (array-like) and scaling factor (float)
    '''
    orig_out = points_along_lines(orig_data, centerlines) # get values along all centerlines
    smooth_out = points_along_lines(smooth_data, centerlines)
    orig_vals = np.array(list(itertools.chain.from_iterable(orig_out[0]))) # combine centerline values to one array
    smooth_vals = np.array(list(itertools.chain.from_iterable(smooth_out[0])))
    
    orig_vals_mean = np.mean(orig_vals) # find the mean of all centerline values
    smooth_vals_mean = np.mean(smooth_vals)
    scaling = orig_vals_mean/smooth_vals_mean # scaling factor between smooth and raw data based on centerline mean
    smooth_data_scaled = rasterio.open(smooth_data).read(1)*scaling
    return smooth_data_scaled, scaling
    
def distance_scaling_correction(data_vals1, data_vals2, distance, outline=None, polyfit_order=2, filter_std=None):
    '''
    Correct smoothed data products based on a scaling from relative distance from centerline
    :param data_vals1: input array/data product 1, e.g. raw data (array-like)
    :param data_vals2: input data product 2--product to be corrected, e.g. smoothed data (array-like)
    :param distance: relative distance of every point from feature(s), e.g. centerlines (array-like)
    :param outline: Boolean glacier outline array (array-like)
    :param polyfit_order: order of the polynomial used to fit data, default 2 (int)
    :param filter_std: filtering outliers based on st.dev, default None (numeric)
    :return:
    '''
    x = distance.flatten()
    y_ratio = np.where(data_vals2 == 0, 1, data_vals1/data_vals2)
    # y_ratio = np.divide(data_vals1, data_vals2)
    y = y_ratio.flatten()
    
    if outline is None:  # if no outline given, consider all array values
    	outline = np.ones_like(distance)
    out = outline.flatten()
    y[out == 0] = -1
    x[out == 0] = -1

    # Filter outliers, if desired
    if filter_std != None:
        mean = np.nanmean(y)
        std = np.nanstd(y)
        y[y > mean + (3 * std)] = np.nan
        x[y > mean + (3 * std)] = np.nan

    # Find binned average values
    stat, edges, __ = binned_statistic(x, y, statistic=np.nanmean, bins=30, range=(max(0, int(np.nanmin(x))), int(np.nanmax(x) + 1)))
    bin_centers = edges[1:] - (abs(edges[0] - edges[1])) / 2
    
    # Fit polynomial to values and use polynomial to derive values at each point
    polyfit = np.polyfit(bin_centers[1:], stat[1:], polyfit_order)
    scale_array = 0
    for n in range(polyfit_order+1):
        scale_array += (polyfit[::-1][n]*(distance**n))
    
    # Correct input data array
    data_vals2_corr = data_vals2*scale_array
    return data_vals2_corr

